from __future__ import annotations

from typing import Optional, Any

import disruptive.logging as dtlog
import disruptive.requests as dtrequests
import disruptive.events.events as dtevents
import disruptive.outputs as dtoutputs
from disruptive.errors import TransferDeviceError, LabelUpdateError


class Device(dtoutputs.OutputBase):
    """
    Represents Sensors and Cloud Connectors, together referred to as devices.

    When a device response is received, the content is
    unpacked and the related attributes are set.

    Attributes
    ----------
    device_id : str
        Unique device ID.
    project_id : str
        Project in which the device resides.
    display_name : str, None
        Given device name if set through a label with key `name`.
        Otherwise None.
    device_type : str
        :ref:`Device type <device_type_constants>`.
    product_number : str
        The product number of the device.
    labels : dict
        Label keys and values.
    is_emulated : bool
        True if the device is emulated, otherwise False.
    reported : Reported
        Object containing the data from the most recent events.
    raw : dict[str, str]
        Unmodified API response JSON.

    """

    # Constants for the various device types.
    TEMPERATURE: str = 'temperature'
    PROXIMITY: str = 'proximity'
    TOUCH: str = 'touch'
    HUMIDITY: str = 'humidity'
    PROXIMITY_COUNTER: str = 'proximityCounter'
    TOUCH_COUNTER: str = 'touchCounter'
    WATER_DETECTOR: str = 'waterDetector'
    CLOUD_CONNECTOR: str = 'ccon'
    CO2: str = 'co2'
    MOTION: str = 'motion'
    DESK_OCCUPANCY: str = 'deskOccupancy'
    CONTACT: str = 'contact'
    DEVICE_TYPES = [
        TEMPERATURE, PROXIMITY, TOUCH, HUMIDITY,
        PROXIMITY_COUNTER, TOUCH_COUNTER, WATER_DETECTOR,
        CLOUD_CONNECTOR, CO2, MOTION, DESK_OCCUPANCY,
    ]

    def __init__(self, device: dict) -> None:
        """
        Constructs the Device object by unpacking the raw device response.

        Parameters
        ----------
        device : dict
            Unmodified device response dictionary.

        """

        # Inherit from OutputBase parent.
        dtoutputs.OutputBase.__init__(self, device)

        # Unpack attributes from dictionary.
        self.device_id: str = device['name'].split('/')[-1]
        self.project_id: str = device['name'].split('/')[1]
        self.device_type: str = device['type']
        self.labels: dict = device['labels']

        # Set display_name if `name` label key exists.
        self.display_name = None
        if 'name' in self.labels:
            self.display_name = self.labels['name']

        # Determine if the device is an emulator by checking id prefix.
        if self.device_id.startswith('emu') and len(self.device_id) == 23:
            self.is_emulated = True
        else:
            self.is_emulated = False

        # If it exists, set the product number.
        # This is not present for emulated devices.
        self.product_number: str = ''
        if 'productNumber' in device:
            self.product_number = device['productNumber']

        # If it exists, set the reported object.
        self.reported = None
        if 'reported' in device:
            self.reported = Reported(device['reported'])

    @classmethod
    def get_device(cls,
                   device_id: str,
                   project_id: Optional[str] = None,
                   **kwargs: Any,
                   ) -> Device:
        """
        Gets the current state of a single device.

        Parameters
        ----------
        device_id : str
            Unique ID of the target device.
        project_id : str, optional
            Unique ID of the target project.
            If not provided, a wildcard project will be used, resulting
            in a search for the device through all available projects.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        device : Device
            Object representing the target device.

        Examples
        --------
        >>> # Fetch a single device.
        >>> device = dt.Device.get_device('<DEVICE_ID>')

        """

        # If project_id is not given, use wildcard "-".
        if project_id is None:
            project_id = '-'

        # Construct URL
        url = '/projects/{}/devices/{}'.format(project_id, device_id)

        # Return Device object of GET request response.
        return cls(dtrequests.DTRequest.get(url, **kwargs))

    @classmethod
    def list_devices(cls,
                     project_id: str,
                     query: Optional[str] = None,
                     device_ids: Optional[list[str]] = None,
                     device_types: Optional[list[str]] = None,
                     label_filters: Optional[dict[str, str]] = None,
                     order_by: Optional[str] = None,
                     **kwargs: Any,
                     ) -> list[Device]:
        """
        Gets a list of the current state of all devices in a
        project, including emulated devices.

        Parameters
        ----------
        project_id : str
            Unique ID of the target project.
        query : str, optional
            Keyword-based device search device type, labels, and identifiers.
            Does not provide additional capabilities over setting explicit
            parameters available in this method.
        device_ids : list[str], optional
            Specify devices by their unique IDs.
        device_types : list[str], optional
            Filter by :ref:`device types <device_type_constants>`.
        label_filters : dict[str, str], optional
            Specify devices by label keys and values (i.e. {"key": "value"}).
            If only a key is provided (i.e. {"key": ""}), all
            device with that key is fetched.
        order_by : str, optional
            The field name you want to order the response by.
            Referred to using dot notation (i.e. "reported.temperature.value").
            Default order is ascending, but can be flipped by prefixing "-".
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        devices : list[Device]
            List of objects each representing a device.

        Examples
        --------
        >>> # List all devices in a project.
        >>> devices = dt.Device.list_devices(
        ...     project_id='<PROJECT_ID>',
        ... )

        >>> # List all touch, temperature, and proximity sensors
        >>> # in a project that has both the labels "room-number"
        >>> # and "external-id" set to spesific values, then
        >>> # sort by their most recent touch event time.
        >>> devices = dt.Device.list_devices(
        ...     project_id=PROJECT_ID,
        ...     device_types=[
        ...         dt.Device.TOUCH,
        ...         dt.Device.TEMPERATURE,
        ...         dt.Device.PROXIMITY,
        ...     ],
        ...     label_filters={
        ...         'room-number': '99',
        ...         'external-id': '55E21B',
        ...     },
        ...     order_by='reported.touch.updateTime',
        ... )

        """

        # Construct parameters dictionary.
        params: dict = dict()
        if query is not None:
            params['query'] = query
        if device_ids is not None:
            params['device_ids'] = device_ids
        if device_types is not None:
            params['device_types'] = device_types
        if order_by is not None:
            params['order_by'] = order_by

        # Convert label_filters dictionary to list of strings.
        if label_filters is not None:
            labels_list = []
            for key in label_filters:
                labels_list.append(key + '=' + label_filters[key])
            params['label_filters'] = labels_list

        # Return list of Device objects of paginated GET response.
        devices = dtrequests.DTRequest.paginated_get(
            url='/projects/{}/devices'.format(project_id),
            pagination_key='devices',
            params=params,
            **kwargs,
        )
        return [cls(device) for device in devices]

    @staticmethod
    def transfer_devices(device_ids: list[str],
                         source_project_id: str,
                         target_project_id: str,
                         **kwargs: Any,
                         ) -> list[TransferDeviceError]:
        """
        Transfers all specified devices to the target project.

        The caller must have the permissions of either `project.admin` or
        `organization.admin` in both the source- and target project.

        Parameters
        ----------
        device_ids : list[str]
            Unique IDs for the devices to transfer.
        source_project_id : str
            Unique ID of the source project.
        target_project_id : str
            Unique ID of the target project.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        errors : list[TransferDeviceError]
            A list that contains one error object for each device that
            could not be successfully transferred.

        Examples
        --------
        >>> # Transfer a device from one project to another.
        >>> err = dt.Device.transfer_devices(
        ...     device_ids=[
        ...         '<DEVICE_ID_1>',
        ...         '<DEVICE_ID_2>',
        ...     ],
        ...     source_project_id='<SOURCE_PROJECT_ID>',
        ...     target_project_id='<TARGET_PROJECT_ID>',
        ... )

        """

        # Construct list of devices.
        name = 'projects/{}/devices/{}'
        devices = [name.format(source_project_id, xid) for xid in device_ids]

        # Construct request body dictionary.
        body = {
            "devices": devices
        }

        # Sent POST request.
        response = dtrequests.DTRequest.post(
            url='/projects/{}/devices:transfer'.format(
                target_project_id
            ),
            body=body,
            **kwargs,
        )

        # Return any transferErrors found in response.
        return [TransferDeviceError(err) for err in response['transferErrors']]

    @staticmethod
    def set_label(device_id: str,
                  project_id: str,
                  key: str,
                  value: str,
                  **kwargs: Any,
                  ) -> list[LabelUpdateError]:
        """
        Set a label key and value for a single device.

        If a label key already exists, the value is updated.

        Parameters
        ----------
        device_id : str
            Unique ID of the target device.
        project_id : str
            Unique ID of the target project.
        key : str
            Label key to be added.
        value : str
            Label value to be added.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        errors : list[LabelUpdateError]
            A list that contains one error object for each label that
            could not be successfully updated.

        Examples
        --------
        >>> # Add a new `room-number` label to a device.
        >>> err = dt.Device.set_label(
        ...     device_id='<DEVICE_ID>',
        ...     project_id='<PROJECT_ID>',
        ...     key='room-number',
        ...     value='99',
        ... )

        """

        # Construct URL.
        url = '/projects/{}/devices:batchUpdate'.format(project_id)

        # Construct request body dictionary.
        body: dict = dict()
        body['devices'] = ['projects/' + project_id + '/devices/' + device_id]
        body['addLabels'] = {key: value}

        # Sent POST request.
        response = dtrequests.DTRequest.post(url, body=body, **kwargs)

        # Return any batchErrors found in response.
        return [LabelUpdateError(err) for err in response['batchErrors']]

    @staticmethod
    def remove_label(device_id: str,
                     project_id: str,
                     key: str,
                     **kwargs: Any,
                     ) -> list[LabelUpdateError]:
        """
        Remove a label (key and value) from a single device.

        Parameters
        ----------
        device_id : str
            Unique ID of the target device.
        project_id : str
            Unique ID of the target project.
        key : str
            Key of the label to be removed.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        errors : list[LabelUpdateError]
            A list that contains one error object for each label that
            could not be successfully updated.

        Examples
        --------
        >>> # Remove the `room-number` label from a device.
        >>> dt.Device.remove_label(
        ...     device_id='<DEVICE_ID>',
        ...     project_id='<PROJECT_ID>',
        ...     key='room-number',
        ... )

        """

        # Construct URL.
        url = '/projects/{}/devices:batchUpdate'.format(project_id)

        # Construct request body dictionary.
        body: dict = dict()
        body['devices'] = ['projects/' + project_id + '/devices/' + device_id]
        body['removeLabels'] = [key]

        # Sent POST request.
        response = dtrequests.DTRequest.post(url, body=body, **kwargs)

        # Return any batchErrors found in response.
        return [LabelUpdateError(err) for err in response['batchErrors']]

    @staticmethod
    def batch_update_labels(device_ids: list[str],
                            project_id: str,
                            set_labels: Optional[dict[str, str]] = None,
                            remove_labels: Optional[list[str]] = None,
                            **kwargs: Any,
                            ) -> list[LabelUpdateError]:
        """
        Add, update, or remove multiple labels (key and value)
        on multiple devices

        Must provide either `set_labels` or `remove_labels`. If neither are
        provided, a :ref:`BadRequest <bad_request_error>` error will be raised.

        Parameters
        ----------
        device_ids : list[str]
            List of unique IDs for the target devices.
        project_id : str
            Unique ID of target project.
        set_labels : dict[str, str], optional
            Key and value of labels to be added. If a label key
            already exists, the value is updated.
        remove_labels : list[str], optional
            Label keys to be removed.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        errors : list[LabelUpdateError]
            A list that contains one error object for each label that
            could not be successfully updated.

        Raises
        ------
        BadRequest
            If neither `set_labels` nor `remove_labels` is provided.

        Examples
        --------
        >>> # Add 3 new and remove 2 old labels for a few devices.
        >>> err = dt.Device.batch_update_labels(
        ...     device_ids=[
        ...         '<DEVICE_1>',
        ...         '<DEVICE_2>',
        ...         '<DEVICE_3>',
        ...     ],
        ...     project_id='<PROJECT_ID>',
        ...     set_labels={
        ...         'room-number': '99',
        ...         'group': 'favorite',
        ...         'alias': 'cookies',
        ...     },
        ...     remove_labels=[
        ...         'external-id',
        ...         'old-tag',
        ...     ],
        ... )

        """

        # Construct list of devices.
        name = 'projects/{}/devices/{}'
        devices = [name.format(project_id, xid) for xid in device_ids]

        # Construct request body dictionary.
        body: dict = dict()
        body['devices'] = devices
        if set_labels is not None:
            body['addLabels'] = set_labels
        if remove_labels is not None:
            body['removeLabels'] = remove_labels

        # Construct URL.
        url = '/projects/{}/devices:batchUpdate'.format(project_id)

        # Sent POST request.
        response = dtrequests.DTRequest.post(url, body=body, **kwargs)

        # Return any batchErrors found in response.
        return [LabelUpdateError(err) for err in response['batchErrors']]


class Reported(dtoutputs.OutputBase):
    """
    Represents the most recent :ref:`Event Data <eventdata>` of
    each type for a single :ref:`Device <device>`.
    If an event type has never been emitted or is not available to
    the device type, the related attribute is `None`.

    Attributes
    ----------
    touch : Touch, None
        Object representing reported touch event data.
    temperature : Temperature, None
        Object representing reported temperature event data.
    object_present : ObjectPresent, None
        Object representing reported objectPresent event data.
    humidity : Humidity, None
        Object representing reported humidity event data.
    object_present_count : ObjectPresentCount, None
        Object representing reported objectPresentCount event data.
    touch_count : TouchCount, None
        Object representing reported touchCount event data.
    water_present : WaterPresent, None
        Object representing reported waterPresent event data.
    network_status : NetworkStatus, None
        Object representing reported networkStatus event data.
    battery_status : BatteryStatus, None
        Object representing reported batteryStatus event data.
    connection_status : ConnectionStatus, None
        Object representing reported connectionStatus event data.
    ethernet_status : EthernetStatus, None
        Object representing reported ethernetStatus event data.
    cellular_status : CellularStatus, None
        Object representing reported cellularStatus event data.
    co2 : Co2, None
        Object representing reported co2 event data.
    pressure : Pressure, None
        Object representing reported pressure event data.
    desk_occupancy : DeskOccupancy, None
        Object representing reported deskOccupancy event data.

    """

    def __init__(self, reported: dict) -> None:
        """
        Constructs the Reported object by unpacking each event in the field.

        Parameters
        ----------
        reported : dict
            Dictionary of the reported field for a device.

        """

        # Inherit parent Event class init.
        dtoutputs.OutputBase.__init__(self, reported)

        # Set default attribute values.
        for key in dtevents._EVENTS_MAP._api_names:
            # Skip labelsChanged as it does not exist in reported.
            if key == 'labelsChanged':
                continue

            # Set attribute to None.
            setattr(self, dtevents._EVENTS_MAP._api_names[key].attr_name, None)

        # Unpack the reported dictionary data.
        self.__unpack()

    def __unpack(self) -> None:
        """
        Iterates each field in the raw dictionary and set the
        related attribute to the appropriate _EventData child object.
        If an event type is not found, the attribute is left as None.

        """

        # Iterate keys in reported dictionary.
        for key in self._raw.keys():
            # Fields can be None on emulated devices. Skip if that is the case.
            if self._raw[key] is None:
                continue

            # Repack the data field in expected format.
            repacked = {key: self._raw[key]}

            # Check if key exists in map of known events.
            if key in dtevents._EVENTS_MAP._api_names:
                # Initialize appropriate data instance.
                data = dtevents._EventData.from_event_type(repacked, key)

                # Set attribute according to event type.
                setattr(
                    self,
                    dtevents._EVENTS_MAP._api_names[key].attr_name,
                    data,
                )
            else:
                dtlog.warning('Skipping unknown reported type {}.'.format(key))



from datetime import datetime

import disruptive.transforms as dttrans


class OutputBase(object):
    """
    Represents common features for all returnable objects.

    Attributes
    ----------
    raw : dict[str, str]
        Unmodified API response JSON.

    """

    def __init__(self, raw: dict) -> None:
        """
        Constructs the OutputBase object by setting raw attribute.

        Parameters
        ----------
        raw : dict[str, str]
            Unmodified API response JSON.

        """

        # Set attribute from input argument.
        self._raw = raw
        self.raw: dict = raw

    def __repr__(self) -> str:
        return '{}.{}({})'.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self._raw,
        )

    def __str__(self) -> str:
        out = self.__str__recursive([], self, level=0)
        return '\n'.join(out)

    def __str__recursive(self, out: list, obj: object, level: int) -> list:
        # Set the indent level for formatting.
        n_spaces = 4
        l0 = level*' '*n_spaces
        l1 = (level+1)*' '*n_spaces
        l2 = (level+2)*' '*n_spaces

        # At first recursive depth, print object name.
        if level == 0:
            out.append(l0 + str(obj.__class__.__name__) + '(')

        # Append the various public attributes recursively.
        for a in vars(obj):
            # Skip private attributes.
            if a.startswith('_'):
                continue
            # Also skip raw as it would only be duplicate information.
            elif a == 'raw':
                continue

            # Fetch and evaluate the attribute value / type.
            val = getattr(obj, a)

            # Class objects should be dumped recursively, except for
            # those that are an instance of datetime, like pandas timestamps.
            if hasattr(val, '__dict__') and not isinstance(val, datetime):
                # Other classes should print name with content recursively.
                out.append('{}{}: {} = {}'.format(
                    l1, a, type(val).__name__,
                    str(val.__class__.__name__) + '('))
                self.__str__recursive(out, val, level=level+1)

            # Lists content should be iterated through.
            elif isinstance(val, list):
                out.append('{}{}: {} = {}'.format(
                    l1, a, type(val).__name__, '['))
                self.__str__list(out, val, level+1, l2)
                out.append(l1 + '],')

            # Other types can be printed directly.
            else:
                out.append('{}{}: {} = {},'.format(
                    l1, a, type(val).__name__, str(val)
                ))

        # At the end of each recursive depth, end object paranthesis.
        out.append(l0 + '),')

        return out

    def __str__list(self,
                    out: list,
                    lst: list,
                    level: int,
                    l1: str,
                    ) -> list:
        for val in lst:
            # Class objects should be dumped recursively.
            if hasattr(val, '__dict__'):
                out.append('{}{}'.format(
                    l1, str(val.__class__.__name__) + '('
                ))
                self.__str__recursive(out, val, level=level+2)

            # Everything else can be printed directly.
            else:
                out.append('{}{} = {},'.format(
                    l1, type(val).__name__, str(val)
                ))
        return out


class Member(OutputBase):
    """
    Represents a member.

    Attributes
    ----------
    member_id : str
        Unique member identifier.
        For Service Account members, this is the Service Account ID.
        For User members, this is the User ID.
    display_name : str
        Provided member display name.
    roles : list[str]
        Roles provided to the member.
    status : str
        Whether the member invite is ACCEPTED or PENDING.
    email : str
        Member email address.
    account_type : str
        Whether the member is a USER or SERVICE_ACCOUNT.
    create_time : datetime
        Timestamp of when the member was created.
    raw : dict[str, str]
        Unmodified API response JSON.

    """

    def __init__(self, member: dict) -> None:
        """
        Constructs the Member object by unpacking the raw response.

        """

        # Inherit from Response parent.
        OutputBase.__init__(self, member)

        # Unpack attributes from dictionary.
        self.member_id = member['name'].split('/')[-1]
        self.display_name = member['displayName']
        self.roles = [r.split('/')[-1] for r in member['roles']]
        self.status = member['status']
        self.email = member['email']
        self.account_type = member['accountType']
        self.create_time = dttrans.to_datetime(member['createTime'])



from __future__ import annotations

from typing import Any, Optional

import requests

import disruptive.outputs as dtoutputs
import disruptive.logging as dtlog


# ------------------------- Parent -------------------------
class DTApiError(Exception):
    """
    Represents errors raised from the REST API.

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)

        # Log the error.
        dtlog.error(message)


# ------------------------- BatchErrors -------------------------
class BatchError(dtoutputs.OutputBase):
    """
    Parent class for errors in batch-style methods where one or several
    actions may fail in an otherwise successful request.

    """

    def __init__(self, error: Any) -> None:
        # Inherit from OutputBase parent.
        super().__init__(error)

        # Log the error.
        dtlog.error(error)


class TransferDeviceError(BatchError):
    """
    Represents errors that occur when a device can, for some reason, not
    be transferred from one project to another.

    Attributes
    ----------
    device_id : str
        Unique ID of the source device.
    project_id : str
        Unique ID of the source project.
    status_code : str
        A status code for the returned error. Is either
        "INVALID_ARGUMENT", "NOT_FOUND", or "INTERNAL_ERROR".
    message : str
        Described the cause of the error.

    """

    def __init__(self, error: dict) -> None:
        # Inherit from OutputBase parent.
        super().__init__(error)

        # Unpack error dictionary.
        self.device_id: str = error['device'].split('/')[-1]
        self.project_id: str = error['device'].split('/')[1]
        self.status_code: str = error['status']['code']
        self.message: str = error['status']['message']


class LabelUpdateError(BatchError):
    """
    Represents errors that occur when a label can, for some reason, not
    be updated for a device.

    Attributes
    ----------
    device_id : str
        Unique ID of the source device.
    project_id : str
        Unique ID of the source project.
    status_code : str
        A status code for the returned error. Is either
        "INVALID_ARGUMENT", "NOT_FOUND", or "INTERNAL_ERROR".
    message : str
        Described the cause of the error.

    """

    def __init__(self, error: dict) -> None:
        # Inherit from OutputBase parent.
        super().__init__(error)

        # Unpack error dictionary.
        self.device_id: str = error['device'].split('/')[-1]
        self.project_id: str = error['device'].split('/')[1]
        self.status_code: str = error['status']['code']
        self.message: str = error['status']['message']


# ------------------------- ServerError -------------------------
class ServerError(DTApiError):
    """
    Covers response errors in the 500 status code range.

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


class InternalServerError(ServerError):
    """
    The response contained a status code of 500.
    https://developer.d21s.com/docs/error-codes#500

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


# ------------------------- UsageError -------------------------
class UsageError(DTApiError):
    """
    Covers response errors in the 400 status code range in
    addition to problems caused by invalid parameter inputs.

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


class BadRequest(UsageError):
    """
    The response contained a status code of 400.
    https://developer.d21s.com/docs/error-codes#400

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


class Unauthorized(UsageError):
    """
    The response contained a status code of 401.
    https://developer.d21s.com/docs/error-codes#401

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


class Forbidden(UsageError):
    """
    The response contained a status code of 403.
    https://developer.d21s.com/docs/error-codes#403

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


class NotFound(UsageError):
    """
    The response contained a status code of 404.
    https://developer.d21s.com/docs/error-codes#404

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


class Conflict(UsageError):
    """
    The response contained a status code of 409.
    https://developer.d21s.com/docs/error-codes#409

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


class TooManyRequests(UsageError):
    """
    The response contained a status code of 429.
    https://developer.d21s.com/docs/error-codes#429

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


class FormatError(UsageError):
    """
    Some string formatting, usually a timestamp, is wrong.

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


class ConfigurationError(UsageError):
    """
    One or more :ref:`configuration parameters <configuration>` are invalid.

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


# ------------------------- ConnectionError -------------------------
class ConnectionError(DTApiError):
    """
    Covers errors caused by unsuccessful connections from the client.
    These are mostly captured requests exceptions.

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


class ReadTimeout(ConnectionError):
    """
    Could not connection to server in the alloted amount
    of time set by :ref:`request_timeout <config params>`.

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


# ------------------------- UnknownError -------------------------
class UnknownError(DTApiError):
    """
    An unexpected exception has been raised.

    This is likely due to a bug in the package.
    Please report this to developer-support@disruptive-technologies.com.

    """

    def __init__(self, message: str | dict) -> None:
        super().__init__(message)


# -------------------------- ClaimErrors ---------------------------
class ClaimError(DTApiError):
    """
    Covers errors related to claiming functions.
    These are mostly captured requests exceptions.

    """

    def __init__(self, error: dict) -> None:
        super().__init__(error)

        # Unmodified API json response.
        self.raw = error


class ClaimErrorDeviceAlreadyClaimed(ClaimError):
    """
    The devices has already been claimed.

    """

    def __init__(self, error: dict) -> None:
        # Inherit from ClaimError parent.
        super().__init__(error)

        self.device_id: str = error['deviceId']
        self.code: str = error['code']
        self.message: str = error['message']


class ClaimErrorKitNotFound(ClaimError):
    """
    The kit was not found.

    """

    def __init__(self, error: dict) -> None:
        # Inherit from ClaimError parent.
        super().__init__(error)

        self.kit_id: str = error['kitId']
        self.code: str = error['code']
        self.message: str = error['message']


class ClaimErrorDeviceNotFound(ClaimError):
    """
    The device was not found.

    """

    def __init__(self, error: dict) -> None:
        # Inherit from ClaimError parent.
        super().__init__(error)

        self.device_id: str = error['deviceId']
        self.code: str = error['code']
        self.message: str = error['message']


# ------------------------- error handling -------------------------
def parse_request_error(caught_error: Exception,
                        data: dict,
                        nth_attempt: int,
                        ) -> tuple:
    """
    Depending on the request error caught, choose a course of action.

    Parameters
    ----------
    caught_error : Exception
        Request error that has been caught.
    data : dict
        Data contained in the error-ridden request.
    nth_attempt : int
        Current request attempt.

    Returns
    -------
    error : Exception
        Exception to be raised.
    should_retry : bool
        If the request should be retried or not.
    nth_attempt : int
        Iterated attempt count.

    """

    # Read Timeouts should be attempted again.
    if isinstance(caught_error, requests.exceptions.ReadTimeout):
        return (
            ReadTimeout('Connection timed out.'),
            True,
            nth_attempt**2,
        )

    # Connection errors should be attempted again.
    elif isinstance(caught_error, requests.exceptions.ConnectionError):
        return (
            ConnectionError('Failed to establish connection.'),
            True,
            nth_attempt**2,
        )
    else:
        # Unhandled error has been raised.
        return caught_error, False, None


def parse_api_status_code(status_code: Optional[int],
                          data: dict,
                          headers: Any,
                          nth_attempt: int,
                          ) -> Any:
    """
    Depending on the status code, returns an exception, retry boolean
    and, incemented retry attempt.

    Parameters
    ----------
    status_code : int
        Status code returned by the API.
    data : dict
        Data in the response.
    headers : dict
        Headers in the response.
    nth_attempt : int
        Retry attempt count.

    Returns
    -------
    error : Exception
        The exception to be raised.
    should_retry : bool
        If the request should be retried.
    nth_attempt : int
        Incremented retry attempt count.

    """

    # If not status code is provided, assume it *couldn't* be
    # set due to an internal error.
    if status_code is None:
        status_code = 500

    # Check for API errors.
    if status_code < 100:
        return InternalServerError(data), True, nth_attempt**2
    elif status_code == 200:
        return None, False, None
    elif status_code == 400:
        return BadRequest(data), False, None
    elif status_code == 401:
        # The first retry is #1. Therefor, retry_count < 2 will
        # result in a a single retry attempt.
        return Unauthorized(data), nth_attempt < 2, None
    elif status_code == 403:
        return Forbidden(data), False, None
    elif status_code == 404:
        return NotFound(data), False, None
    elif status_code == 409:
        return Conflict(data), False, None
    elif status_code == 429:
        if 'Retry-After' in headers:
            return (
                TooManyRequests(data),
                True,
                int(headers['Retry-After']),
            )
        else:
            return TooManyRequests(data), False, None
    elif status_code == 500:
        return InternalServerError(data), True, nth_attempt**2
    elif status_code == 503:
        return InternalServerError(data), True, nth_attempt**2
    elif status_code == 504:
        return InternalServerError(data), True, nth_attempt**2 + 9
    else:
        return UnknownError(data), False, None


def _raise_builtin(error: Any, message: str) -> Any:
    """
    Used to accompany raised builtin errors with a log message.

    Parameters
    ----------
    error : Exception
        Exception to be raised after logging.
    message : str
        Message to log.

    """

    # Log the error.
    dtlog.error(message)

    # Raise provided error.
    raise error(message)



from __future__ import annotations

from typing import Optional, Union
from datetime import datetime

import disruptive
import disruptive.logging as dtlog
import disruptive.outputs as dtoutputs
import disruptive.transforms as dttrans

TOUCH: str = 'touch'
TEMPERATURE: str = 'temperature'
OBJECT_PRESENT: str = 'objectPresent'
HUMIDITY: str = 'humidity'
OBJECT_PRESENT_COUNT: str = 'objectPresentCount'
TOUCH_COUNT: str = 'touchCount'
WATER_PRESENT: str = 'waterPresent'
NETWORK_STATUS: str = 'networkStatus'
BATTERY_STATUS: str = 'batteryStatus'
LABELS_CHANGED: str = 'labelsChanged'
CONNECTION_STATUS: str = 'connectionStatus'
ETHERNET_STATUS: str = 'ethernetStatus'
CELLULAR_STATUS: str = 'cellularStatus'
CO2: str = 'co2'
PRESSURE: str = 'pressure'
MOTION: str = 'motion'
DESK_OCCUPANCY: str = 'deskOccupancy'
CONTACT: str = 'contact'
PROBE_WIRE_STATUS: str = 'probeWireStatus'
EVENT_TYPES = [
    TOUCH, TEMPERATURE, OBJECT_PRESENT, HUMIDITY, OBJECT_PRESENT_COUNT,
    TOUCH_COUNT, WATER_PRESENT, NETWORK_STATUS, BATTERY_STATUS, LABELS_CHANGED,
    CONNECTION_STATUS, ETHERNET_STATUS, CELLULAR_STATUS, CO2, PRESSURE, MOTION,
    DESK_OCCUPANCY, CONTACT, PROBE_WIRE_STATUS,
]


class _EventData(dtoutputs.OutputBase):
    """
    Parent class for all the different event data field types.

    Attributes are mainly inherited from OutputBase, and otherwise
    contains a few convenience methods for setting the correct child.

    Attributes
    ----------
    timestamp : datetime, str, optional
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self, data: dict, event_type: str) -> None:
        """
        Constructs the _EventData object by inheriting parent.

        Parameters
        ----------
        data : dict
            Event data dictionary.
        event_type : str
            Name of the event type.

        """

        self.timestamp: Optional[datetime | str] = None

        # If timestamp is provided, verify type and set attribute.
        if 'updateTime' in data:
            # Raw should be iso8601 str format, while
            # attribute should be type datetime.
            ts_iso8601 = dttrans.to_iso8601(data['updateTime'])
            ts_datetime = dttrans.to_datetime(data['updateTime'])

            # If we can not verify iso8601 format, remove field.
            if ts_iso8601 is not None:
                data['updateTime'] = ts_iso8601
            else:
                del data['updateTime']

            # Set datetime return as timestamp attribute.
            self.timestamp = ts_datetime
        else:
            # None timestamp if not provided.
            self.timestamp = None

        # Set other attributes.
        self.event_type = event_type

        # Inherit parent class.
        dtoutputs.OutputBase.__init__(self, data)

    @classmethod
    def from_event_type(cls,
                        data: dict,
                        event_type: str,
                        ) -> Optional[_EventType]:
        """
        Constructs the appropriate child class from the provided event type.

        Parameters
        ----------
        data : dict
            Dictionary of the data field found in the source event.
        event_type : str
            The event type of the source event.

        """

        # From the type, select the appropriate child class.
        child, key = cls.__child_map(event_type)

        # Return None at invalid event type.
        if child is None:
            return None

        # Return the initialized class instance.
        if key:
            child_instance = child._from_raw(data[event_type])
        else:
            # Special case for labelsChanged event.
            child_instance = child._from_raw(data)
        return child_instance

    @staticmethod
    def __child_map(event_type: str,
                    ) -> tuple[Optional[_EventType], Optional[bool]]:
        """
        Based on provided event type, returns the
        child class and supporting information.

        Parameters
        ----------
        event_type : str
            The event type of the source event.

        """

        # Initialize the correct object.
        if event_type in _EVENTS_MAP._api_names:
            out = (
                getattr(
                    disruptive.events,
                    _EVENTS_MAP._api_names[event_type].class_name
                ),
                _EVENTS_MAP._api_names[event_type].is_keyed,
            )
            return out

        dtlog.warning('Skipping unknown event type {}.'.format(event_type))
        return None, None


class Touch(_EventData):
    """
    Represents the data found in a touch event.

    Attributes
    ----------
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self, timestamp: Optional[datetime | str] = None):
        """
        Constructs the Touch object.

        Parameters
        ----------
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'touch')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> Touch:
        """
        Constructs a Touch object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : Touch
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class Temperature(_EventData):
    """
    Represents the data found in a temperature event.

    Attributes
    ----------
    celsius : float
        Temperature value in Celsius.
    fahrenheit : float
        Temperature value in Fahrenheit.
    samples : list[TemperatureSample]
        Temperature values sampled over a single heartbeat.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self,
                 celsius: float,
                 samples: Optional[list] = None,
                 timestamp: Optional[datetime | str] = None,
                 ) -> None:
        """
        Constructs the Temperature object. The `fahrenheit` attribute is
        calculated from the provided `celsius` parameter.

        Parameters
        ----------
        celsius : float
            Temperature value in Celsius.
        samples : list[TemperatureSample]
            Temperature values sampled over a single heartbeat.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.celsius: float = celsius
        self.samples: Optional[list] = samples
        self.fahrenheit: float = dttrans._celsius_to_fahrenheit(celsius)
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'temperature')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'celsius={}, '\
            'samples={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.celsius,
            self.samples,
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> Temperature:
        """
        Constructs a Temperature object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : Temperature
            Object constructed from the API response data.

        """

        # Convert samples dictionaries to TemperatureSample objects.
        sample_objs = []
        for sample in data['samples']:
            sample_objs.append(TemperatureSample(
                celsius=sample['value'],
                timestamp=sample['sampleTime'],
            ))

        # Construct the object with unpacked parameters.
        obj = cls(
            celsius=data['value'],
            samples=sample_objs,
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.celsius is not None:
            data['value'] = self.celsius
        if self.samples is not None:
            data['samples'] = [s._raw for s in self.samples]
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class TemperatureSample(dtoutputs.OutputBase):
    """
    Represents a single temperature event sample from a heartbeat period.

    Attributes
    ----------
    celsius : float
        Temperature value in Celsius.
    fahrenheit : float
        Temperature value in Fahrenheit.
    timestamp : datetime
        Interpolated inter-heartbeat timestamp.

    """

    def __init__(self,
                 celsius: float,
                 timestamp: datetime | str,
                 ) -> None:
        """
        Constructs the TemperatureSample object. The `fahrenheit` attribute is
        calculated from the provided `celsius` parameter.

        Parameters
        ----------
        celsius : float
            Temperature value in Celsius.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.celsius: float = celsius
        self.fahrenheit: float = dttrans._celsius_to_fahrenheit(celsius)
        self.timestamp = dttrans.to_datetime(timestamp)

        # Inherit parent class.
        dtoutputs.OutputBase.__init__(self, self.__repack())

    def __repr__(self) -> str:
        string = '{}.{}('\
            'celsius={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.celsius,
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> TemperatureSample:
        # Construct the object with unpacked parameters.
        obj = cls(
            celsius=data['value'],
            timestamp=data['sampleTime'],
        )

        # Inherit parent class.
        dtoutputs.OutputBase.__init__(obj, data)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.celsius is not None:
            data['value'] = self.celsius
        if self.timestamp is not None:
            data['sampleTime'] = dttrans.to_iso8601(self.timestamp)
        return data


class ObjectPresent(_EventData):
    """
    Represents the data found in an objectPresent event.

    Attributes
    ----------
    state : str
        Indicates whether an object is "PRESENT" or "NOT_PRESENT".
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    STATE_PRESENT = 'PRESENT'
    STATE_NOT_PRESENT = 'NOT_PRESENT'

    def __init__(self,
                 state: str,
                 timestamp: Optional[datetime | str] = None,
                 ) -> None:
        """
        Constructs the ObjectPresent object, inheriting parent class
        and setting the type-specific attributes.

        Parameters
        ----------
        state : str
            Indicates whether an object is "PRESENT" or "NOT_PRESENT".
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.state: str = state
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'objectPresent')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'state={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            repr(self.state),
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> ObjectPresent:
        """
        Constructs an ObjectPresent object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : ObjectPresent
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            state=data['state'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.state is not None:
            data['state'] = self.state
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class Humidity(_EventData):
    """
    Represents the data found in an humidity event.

    Attributes
    ----------
    temperature : float
        Temperature value in Celsius.
    relative_humidity : float
        Relative humidity in percent.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self,
                 celsius: float,
                 relative_humidity: float,
                 timestamp: Optional[datetime | str] = None,
                 ):
        """
        Constructs the Humidity object.

        Parameters
        ----------
        temperature : float
            Temperature value in Celsius.
        relative_humidity : float
            Relative humidity in percent.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.celsius: float = celsius
        self.fahrenheit: float = dttrans._celsius_to_fahrenheit(celsius)
        self.relative_humidity: float = relative_humidity
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'humidity')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'celsius={}, '\
            'relative_humidity={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.celsius,
            self.relative_humidity,
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> Humidity:
        """
        Constructs a Humidity object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : Humidity
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            celsius=data['temperature'],
            relative_humidity=data['relativeHumidity'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.celsius is not None:
            data['temperature'] = self.celsius
        if self.relative_humidity is not None:
            data['relativeHumidity'] = self.relative_humidity
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class ObjectPresentCount(_EventData):
    """
    Represents the data found in an objectPresentCount event.

    Attributes
    ----------
    total : int
        The total number of times the sensor has detected the appearance
        or disappearance of an object over its lifetime.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self,
                 total: int,
                 timestamp: Optional[datetime | str] = None,
                 ):
        """
        Constructs the ObjectPresentCount object.

        Parameters
        ----------
        total : int
            The total number of times the sensor has detected the appearance
            or disappearance of an object over its lifetime.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.total: int = total
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'objectPresentCount')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'total={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.total,
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> ObjectPresentCount:
        """
        Constructs a ObjectPresentCount object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : ObjectPresentCount
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            total=data['total'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        data['total'] = self.total
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class TouchCount(_EventData):
    """
    Represents the data found in an touchCount event.

    Attributes
    ----------
    total : int
        The total number of times the sensor has been
        touched over its lifetime.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self,
                 total: int,
                 timestamp: Optional[datetime | str] = None,
                 ):
        """
        Constructs the TouchCount object.

        Parameters
        ----------
        total : int
            The total number of times the sensor has been
            touched over its lifetime.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.total: int = total
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'touchCount')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'total={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.total,
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> TouchCount:
        """
        Constructs a TouchCount object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : TouchCount
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            total=data['total'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        data['total'] = self.total
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class WaterPresent(_EventData):
    """
    Represents the data found in an waterPresent event.

    Attributes
    ----------
    state : str
        Indicates whether water is PRESENT or NOT_PRESENT.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    STATE_PRESENT = 'PRESENT'
    STATE_NOT_PRESENT = 'NOT_PRESENT'

    def __init__(self,
                 state: str,
                 timestamp: Optional[datetime | str] = None,
                 ):
        """
        Constructs the WaterPresent object.

        Parameters
        ----------
        state : str
            Indicates whether water is "PRESENT" or "NOT_PRESENT".
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.state: str = state
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'waterPresent')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'state={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            repr(self.state),
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> WaterPresent:
        """
        Constructs a WaterPresent object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : WaterPresent
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            state=data['state'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        data['state'] = self.state
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class NetworkStatusCloudConnector(dtoutputs.OutputBase):
    """
    Represents a Cloud Connector found in the networkStatus event data.

    Attributes
    ----------
    device_id : str
        Device ID of the Cloud Connector.
    signal_strength : int
        The percentage signal strength (0% to 100%) between
        the sensor and Cloud Connector.
    rssi : int
        Raw Received Signal Strength Indication (RSSI) between
        the sensor and Cloud Connector.

    """

    def __init__(self,
                 device_id: str,
                 signal_strength: int,
                 rssi: int,
                 ):
        """
        Constructs the NetworkStatusCloudConnector object.

        Parameters
        ----------
        device_id : str
            Device ID of the Cloud Connector.
        signal_strength : int
            The percentage signal strength (0% to 100%) between
            the sensor and Cloud Connector.
        rssi : int
            Raw Received Signal Strength Indication (RSSI) between
            the sensor and Cloud Connector.

        """

        # Inherit parent OutputBase class init.
        dtoutputs.OutputBase.__init__(self, {})

        # Unpack attributes.
        self.device_id: str = device_id
        self.signal_strength: int = signal_strength
        self.rssi: int = rssi

    def __repr__(self) -> str:
        string = '{}.{}('\
            'device_id={}, '\
            'signal_strength={}, '\
            'rssi={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            repr(self.device_id),
            self.signal_strength,
            self.rssi,
        )

    @classmethod
    def _from_raw(cls, data: dict) -> NetworkStatusCloudConnector:
        """
        Constructs a NetworkStatusCloudConnector object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : NetworkStatusCloudConnector
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            device_id=data['id'],
            signal_strength=data['signalStrength'],
            rssi=data['rssi'],
        )

        # Re-inherit from parent, but now providing response data.
        dtoutputs.OutputBase.__init__(obj, data)

        return obj


class NetworkStatus(_EventData):
    """
    Represents the data found in a networkStatus event.

    Attributes
    ----------
    signal_strength : int
        The percentage signal strength (0% to 100%) of the strongest
        Cloud Connector, derived directly from the RSSI value.
    rssi : float
        Raw Received Signal Strength Indication (RSSI) as measured
        by the strongest Cloud Connector.
    transmission_mode : str
        Indicated whether the sensor is in
        LOW_POWER_STANDARD_MODE or HIGH_POWER_BOOST_MODE.
    cloud_connectors : list[NetworkStatusCloudConnector]
        List of the Cloud Connector that forwarded the event.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self,
                 signal_strength: Optional[int] = None,
                 rssi: Optional[int] = None,
                 transmission_mode: Optional[str] = None,
                 cloud_connectors:
                 Optional[list[NetworkStatusCloudConnector]] = None,
                 timestamp: Optional[datetime | str] = None,
                 ) -> None:

        """
        Constructs the NetworkStatus object.

        Parameters
        ----------
        signal_strength : int
            The percentage signal strength (0% to 100%) of the strongest
            Cloud Connector, derived directly from the RSSI value.
        rssi : float
            Raw Received Signal Strength Indication (RSSI) as measured
            by the strongest Cloud Connector.
        transmission_mode : str
            Indicated whether the sensor is in
            LOW_POWER_STANDARD_MODE or HIGH_POWER_BOOST_MODE.
        cloud_connectors : list[NetworkStatusCloudConnector]
            List of the Cloud Connector that forwarded the event.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set attributes.
        self.signal_strength: Optional[int] = signal_strength
        self.rssi: Optional[int] = rssi
        self.transmission_mode: Optional[str] = transmission_mode
        self.cloud_connectors: Optional[list[NetworkStatusCloudConnector]] = \
            cloud_connectors
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'networkStatus')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'signal_strength={}, '\
            'rssi={}, '\
            'transmission_mode={}, '\
            'cloud_connectors={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.signal_strength,
            self.rssi,
            repr(self.transmission_mode),
            self.cloud_connectors,
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> NetworkStatus:
        """
        Constructs a NetworkStatus object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : NetworkStatus
            Object constructed from the API response data.

        """

        # Isolate list of NetworkStatusCloudConnector objects.
        cloud_connectors = []
        for ccon in data['cloudConnectors']:
            cloud_connectors.append(
                NetworkStatusCloudConnector._from_raw(ccon)
            )

        # Construct the object with unpacked parameters.
        obj = cls(
            signal_strength=data['signalStrength'],
            rssi=data['rssi'],
            transmission_mode=data['transmissionMode'],
            cloud_connectors=cloud_connectors,
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.signal_strength is not None:
            data['signalStrength'] = self.signal_strength
        if self.rssi is not None:
            data['rssi'] = self.rssi
        if self.transmission_mode is not None:
            data['transmissionMode'] = self.transmission_mode
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        if self.cloud_connectors is not None:
            data['cloud_connectors'] = []
            for ccon in self.cloud_connectors:
                ccon_data: dict = dict()
                if ccon.device_id is not None:
                    ccon_data['id'] = ccon.device_id
                if ccon.signal_strength is not None:
                    ccon_data['signalStrength'] = ccon.signal_strength
                if len(ccon_data) > 0:
                    data['cloud_connectors'].append(ccon_data)
        return data


class BatteryStatus(_EventData):
    """
    Represents the data found in a batteryStatus event.

    Attributes
    ----------
    percentage : int
        A coarse percentage estimate (0% to 100%) of the remaining battery.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self,
                 percentage: int,
                 timestamp: Optional[datetime | str] = None,
                 ) -> None:
        """
        Constructs the Temperature object.

        Parameters
        ----------
        percentage : int
            A coarse percentage estimate (0% to 100%) of the remaining battery.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Inherit parent _EventData class init with repacked data dictionary.
        self.percentage: int = percentage
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'batteryStatus')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'percentage={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.percentage,
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> BatteryStatus:
        """
        Constructs a BatteryStatus object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : BatteryStatus
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            percentage=data['percentage'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.percentage is not None:
            data['percentage'] = self.percentage
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class LabelsChanged(_EventData):
    """
    Represents the data found in an labelsChanged event.

    Attributes
    ----------
    added : dict[str, str]
        Keys and values of new labels added.
    modified : dict[str, str]
        New keys and values of modified labels.
    removed : list[str]
        List of keys of removed labels.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self,
                 added: dict[str, str],
                 modified: dict[str, str],
                 removed: list[str],
                 timestamp: Optional[datetime | str] = None,
                 ):
        """
        Constructs the LabelsChanged object.

        Parameters
        ----------
        added : dict[str, str]
            Keys and values of new labels added.
        modified : dict[str, str]
            New keys and values of modified labels.
        removed : list[str]
            List of keys of removed labels.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.added: dict = added
        self.modified: dict = modified
        self.removed: list[str] = removed
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'labelsChanged')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'added={}, '\
            'modified={}, '\
            'removed={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.added,
            self.modified,
            self.removed,
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> LabelsChanged:
        """
        Constructs a LabelsChanged object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : LabelsChanged
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            added=data['added'],
            modified=data['modified'],
            removed=data['removed'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.added is not None:
            data['added'] = self.added
        if self.modified is not None:
            data['modified'] = self.modified
        if self.removed is not None:
            data['removed'] = self.removed
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class ConnectionStatus(_EventData):
    """
    Represents the data found in a connectionStatus event.

    Attributes
    ----------
    connection : str
        Whether the Cloud Connector is on "ETHERNET", "CELLULAR", or "OFFLINE".
    available : list[str]
        Lists available connections. Can contain
        "ETHERNET", "CELLULAR", or both.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self,
                 connection: str,
                 available: list[str],
                 timestamp: Optional[datetime | str] = None,
                 ):
        """
        Constructs the ConnectionStatus object.

        Parameters
        ----------
        connection : str
            Whether the Cloud Connector is on
            "ETHERNET", "CELLULAR", or "OFFLINE".
        available : list[str]
            Lists available connections.
            Can contain "ETHERNET", "CELLULAR", or both.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.connection: str = connection
        self.available: list[str] = available
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'connectionStatus')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'connection={}, '\
            'available={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            repr(self.connection),
            repr(self.available),
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> ConnectionStatus:
        """
        Constructs a ConnectionStatus object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : ConnectionStatus
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            connection=data['connection'],
            available=data['available'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.connection is not None:
            data['connection'] = self.connection
        if self.available is not None:
            data['available'] = self.available
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class EthernetStatus(_EventData):
    """
    Represents the data found in a ethernetStatus event.

    Attributes
    ----------
    mac_address : str
        MAC address of the local network interface.
    ip_address : str
        IP address of the Cloud Connector on the local network.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self,
                 mac_address: str,
                 ip_address: str,
                 timestamp: Optional[datetime | str] = None,
                 ):
        """
        Constructs the EthernetStatus object, inheriting parent class
        and setting the type-specific attributes.

        Parameters
        ----------
        mac_address : str
            MAC address of the local network interface.
        ip_address : str
            IP address of the Cloud Connector on the local network.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.mac_address: str = mac_address
        self.ip_address: str = ip_address
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'ethernetStatus')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'mac_address={}, '\
            'ip_address={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            repr(self.mac_address),
            repr(self.ip_address),
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> EthernetStatus:
        """
        Constructs a EthernetStatus object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : EthernetStatus
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            mac_address=data['macAddress'],
            ip_address=data['ipAddress'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.mac_address is not None:
            data['macAddress'] = self.mac_address
        if self.ip_address is not None:
            data['ipAddress'] = self.ip_address
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class CellularStatus(_EventData):
    """
    Represents the data found in a cellularStatus event.

    Attributes
    ----------
    signal_strength : int
        Cloud Connector cellular reception percentage.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self,
                 signal_strength: int,
                 timestamp: Optional[datetime | str] = None,
                 ):
        """
        Constructs the Temperature object.

        Parameters
        ----------
        signal_strength : int
            Cloud Connector cellular reception percentage.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.signal_strength: int = signal_strength
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'cellularStatus')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'signal_strength={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.signal_strength,
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> CellularStatus:
        """
        Constructs a CellularStatus object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : CellularStatus
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            signal_strength=data['signalStrength'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent using a full raw dictionary.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.signal_strength is not None:
            data['signalStrength'] = self.signal_strength
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class Co2(_EventData):
    """
    Represents the data found in a co2 event.

    Attributes
    ----------
    ppm : int
        Co2 concentration in parts per million.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self,
                 ppm: int,
                 timestamp: Optional[datetime | str] = None,
                 ) -> None:
        """
        Constructs the Co2 object.

        Parameters
        ----------
        ppm : int
            Co2 concentration in parts per million.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.ppm: int = ppm
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'co2')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'ppm={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.ppm,
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> Co2:
        """
        Constructs a Co2 object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : Co2
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            ppm=data['ppm'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.ppm is not None:
            data['ppm'] = self.ppm
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class Pressure(_EventData):
    """
    Represents the data found in a pressure event.

    Attributes
    ----------
    pascal : int
       Barometric pressure in pascal.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    def __init__(self,
                 pascal: float,
                 timestamp: Optional[datetime | str] = None,
                 ) -> None:
        """
        Constructs the Pressure object.

        Parameters
        ----------
        pascal : int
            Barometric pressure in pascal.
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.pascal: float = pascal
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'pressure')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'pascal={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.pascal,
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> Pressure:
        """
        Constructs a Pressure object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : Pressure
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            pascal=data['pascal'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.pascal is not None:
            data['pascal'] = self.pascal
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class Motion(_EventData):
    """
    Represents the data found in a motion event.

    Attributes
    ----------
    state : str
        Indicates whether "MOTION_DETECTED" or "NO_MOTION_DETECTED".
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    STATE_MOTION_DETECTED = 'MOTION_DETECTED'
    STATE_NO_MOTION_DETECTED = 'NO_MOTION_DETECTED'

    def __init__(self,
                 state: str,
                 timestamp: Optional[datetime | str] = None,
                 ) -> None:
        """
        Constructs the Motion object, inheriting parent class
        and setting the type-specific attributes.

        Parameters
        ----------
        state : str
            Indicates whether "MOTION_DETECTED" or "NO_MOTION_DETECTED".
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        # Set parameter attributes.
        self.state: str = state
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'motion')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'state={}, '\
            'timestamp={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            repr(self.state),
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> Motion:
        """
        Constructs a Motion object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : Motion
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            state=data['state'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.state is not None:
            data['state'] = self.state
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class DeskOccupancy(_EventData):
    """
    Represents the data found in a deskOccupancy event.

    Attributes
    ----------
    state : str
        Indicates whether the sensor predicts "OCCUPIED" or "NOT_OCCUPIED".
    remarks : list[str], optional
        Additional information about the estimated state, given as
        a list of string remarks.
        See https://developer.d21s.com/docs/concepts/events#remarks.
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    STATE_OCCUPIED = 'OCCUPIED'
    STATE_NOT_OCCUPIED = 'NOT_OCCUPIED'

    def __init__(self,
                 state: str,
                 timestamp: Optional[datetime | str] = None,
                 remarks: Optional[list[str]] = None,
                 ) -> None:
        """
        Constructs the DeskOccupancy object, inheriting parent class
        and setting the type-specific attributes.

        Parameters
        ----------
        state : str
            Indicates whether the sensor predicts "OCCUPIED" or "NOT_OCCUPIED".
        timestamp : datetime, str, optional
            Timestamp in either datetime or string iso8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).
        remarks : list[str], optional
            Additional information about the estimated state, given as
            a list of string remarks.
            See https://developer.d21s.com/docs/concepts/events#remarks.

        """

        # Set parameter attributes.
        self.state: str = state
        self.timestamp: Optional[datetime | str] = timestamp
        self.remarks: Optional[list[str]] = remarks

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'deskOccupancy')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'state={}, '\
            'timestamp={}, '\
            'remarks={}'\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            repr(self.state),
            repr(dttrans.to_iso8601(self.timestamp)),
            repr(self.remarks),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> DeskOccupancy:
        """
        Constructs a DeskOccupancy object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : DeskOccupancy
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            state=data['state'],
            timestamp=data['updateTime'],
            remarks=data['remarks'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.state is not None:
            data['state'] = self.state
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        if self.remarks is not None:
            data['remarks'] = self.remarks
        return data


class Contact(_EventData):
    """
    Represents the data found in a contact event.

    Attributes
    ----------
    state : str
        Indicates whether the sensor detects "CLOSED" or "OPEN".
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    STATE_CLOSED: str = 'CLOSED'
    STATE_OPEN: str = 'OPEN'

    def __init__(self,
                 state: str,
                 timestamp: Optional[datetime | str] = None,
                 ) -> None:
        """
        Constructs the Contact object, inheriting parent class
        and setting the type-specific attributes.

        Parameters
        ----------
        state : str
            Indicates whether the sensor predicts "CLOSED" or "OPEN".
        timestamp : datetime, str, optional
            Timestamp in either datetime or string ISO8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        self.state: str = state
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'contact')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'state={}, '\
            'timestamp={}, '\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            repr(self.state),
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> Contact:
        """
        Constructs a Contact object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : Contact
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            state=data['state'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.state is not None:
            data['state'] = self.state
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class ProbeWireStatus(_EventData):
    """
    Represents the data found in a contact event.

    Attributes
    ----------
    timestamp : datetime
        Timestamp of when the event was received by a Cloud Connector.

    """

    STATE_INVALID_WIRE_CONFIGURATION: str = 'INVALID_WIRE_CONFIGURATION'
    STATE_TWO_WIRE: str = 'TWO_WIRE'
    STATE_THREE_WIRE: str = 'THREE_WIRE'
    STATE_FOUR_WIRE: str = 'FOUR_WIRE'

    def __init__(self,
                 state: str,
                 timestamp: Optional[datetime | str] = None,
                 ) -> None:
        """
        Constructs the ProbeWireStatus object, inheriting parent class
        and setting the type-specific attributes.

        Parameters
        ----------
        state : str
            Probe wire status. Can be either "INVALID_WIRE_CONFIGURATION",
            "TWO_WIRE", "THREE_WIRE", or "FOUR_WIRE".
        timestamp : datetime, str, optional
            Timestamp in either datetime or string ISO8601 format
            (i.e. yyyy-MM-ddTHH:mm:ssZ).

        """

        self.state: str = state
        self.timestamp: Optional[datetime | str] = timestamp

        # Inherit parent _EventData class init with repacked data dictionary.
        _EventData.__init__(self, self.__repack(), 'probeWireStatus')

    def __repr__(self) -> str:
        string = '{}.{}('\
            'state={}, '\
            'timestamp={}, '\
            ')'
        return string.format(
            self.__class__.__module__,
            self.__class__.__name__,
            repr(self.state),
            repr(dttrans.to_iso8601(self.timestamp)),
        )

    @classmethod
    def _from_raw(cls, data: dict) -> ProbeWireStatus:
        """
        Constructs a ProbeWireStatus object from API response data.

        Parameters
        ----------
        data : dict
            API response data dictionary.

        Returns
        -------
        obj : Contact
            Object constructed from the API response data.

        """

        # Construct the object with unpacked parameters.
        obj = cls(
            state=data['state'],
            timestamp=data['updateTime'],
        )

        # Re-inherit from parent, but now providing response data.
        _EventData.__init__(obj, data, obj.event_type)

        return obj

    def __repack(self) -> dict:
        data: dict = dict()
        if self.state is not None:
            data['state'] = self.state
        if self.timestamp is not None:
            data['updateTime'] = self.timestamp
        return data


class Event(dtoutputs.OutputBase):
    """
    Represents device events.

    Attributes
    ----------
    event_id : str
        Unique event ID.
    event_type : str
        Event type.
    device_id : str
        Unique ID of the source device.
    project_id : str
        Unique ID of the source project.
    data : :ref:`Event Data <eventdata>`
        An object representing type-specific event data.
    raw : dict[str, str]
        Unmodified API response JSON.

    """

    def __init__(self, event: dict):
        # Inherit attributes from ResponseBase parent.
        dtoutputs.OutputBase.__init__(self, event)

        # Unpack attributes from dictionary.
        self.event_id: str = event['eventId']
        self.event_type: str = event['eventType']
        self.device_id: str = event['targetName'].split('/')[-1]
        self.project_id: str = event['targetName'].split('/')[1]

        # Since labelsChanged is the only event that does not
        # contain an updateTime field in data, we provide the
        # field as it is a massive convenience boost.
        if self.event_type == 'labelsChanged':
            event['data']['updateTime'] = event['timestamp']

        # Initialize the appropriate data class.
        self.data = _EventData.from_event_type(
            event['data'],
            self.event_type,
        )

    @classmethod
    def from_mixed_list(cls, events: list[dict]) -> list[Event]:
        """
        Construct Event objects for each event in list.

        Parameters
        ----------
        events : list[dict]
            List of raw event response dictionaries.

        Returns
        -------
        events : list[Event]
            List of constructed event objects.

        """

        # Initialise output list.
        object_list = []

        # Iterate events in list.
        for event in events:
            # Initialize instance and append to output.
            object_list.append(cls(event))

        return object_list


class __EventsMap():

    class __TypeNames():

        def __init__(self,
                     api_name: str,
                     attr_name: str,
                     class_name: str,
                     is_keyed: bool
                     ) -> None:
            self.api_name = api_name
            self.attr_name = attr_name
            self.class_name = class_name
            self.is_keyed = is_keyed

    _api_names = {
        'touch': __TypeNames(
            api_name='touch',
            attr_name='touch',
            class_name='Touch',
            is_keyed=True
        ),
        'temperature': __TypeNames(
            api_name='temperature',
            attr_name='temperature',
            class_name='Temperature',
            is_keyed=True
        ),
        'objectPresent': __TypeNames(
            api_name='objectPresent',
            attr_name='object_present',
            class_name='ObjectPresent',
            is_keyed=True
        ),
        'humidity': __TypeNames(
            api_name='humidity',
            attr_name='humidity',
            class_name='Humidity',
            is_keyed=True
        ),
        'objectPresentCount': __TypeNames(
            api_name='objectPresentCount',
            attr_name='object_present_count',
            class_name='ObjectPresentCount',
            is_keyed=True
        ),
        'touchCount': __TypeNames(
            api_name='touchCount',
            attr_name='touch_count',
            class_name='TouchCount',
            is_keyed=True
        ),
        'waterPresent': __TypeNames(
            api_name='waterPresent',
            attr_name='water_present',
            class_name='WaterPresent',
            is_keyed=True
        ),
        'networkStatus': __TypeNames(
            api_name='networkStatus',
            attr_name='network_status',
            class_name='NetworkStatus',
            is_keyed=True,
        ),
        'batteryStatus': __TypeNames(
            api_name='batteryStatus',
            attr_name='battery_status',
            class_name='BatteryStatus',
            is_keyed=True,
        ),
        'labelsChanged': __TypeNames(
            api_name='labelsChanged',
            attr_name='labels_changed',
            class_name='LabelsChanged',
            is_keyed=False,
        ),
        'connectionStatus': __TypeNames(
            api_name='connectionStatus',
            attr_name='connection_status',
            class_name='ConnectionStatus',
            is_keyed=True,
        ),
        'ethernetStatus': __TypeNames(
            api_name='ethernetStatus',
            attr_name='ethernet_status',
            class_name='EthernetStatus',
            is_keyed=True,
        ),
        'cellularStatus': __TypeNames(
            api_name='cellularStatus',
            attr_name='cellular_status',
            class_name='CellularStatus',
            is_keyed=True,
        ),
        'co2': __TypeNames(
            api_name='co2',
            attr_name='co2',
            class_name='Co2',
            is_keyed=True,
        ),
        'pressure': __TypeNames(
            api_name='pressure',
            attr_name='pressure',
            class_name='Pressure',
            is_keyed=True,
        ),
        'motion': __TypeNames(
            api_name='motion',
            attr_name='motion',
            class_name='Motion',
            is_keyed=True,
        ),
        'deskOccupancy': __TypeNames(
            api_name='deskOccupancy',
            attr_name='desk_occupancy',
            class_name='DeskOccupancy',
            is_keyed=True,
        ),
        'contact': __TypeNames(
            api_name='contact',
            attr_name='contact',
            class_name='Contact',
            is_keyed=True,
        ),
        'probeWireStatus': __TypeNames(
            api_name='probeWireStatus',
            attr_name='probe_wire_status',
            class_name='ProbeWireStatus',
            is_keyed=True,
        ),
    }


_EVENTS_MAP = __EventsMap()

_EventType = Union[
    Touch, Temperature, ObjectPresent, Humidity, ObjectPresentCount,
    TouchCount, WaterPresent, NetworkStatus, NetworkStatusCloudConnector,
    BatteryStatus, LabelsChanged, ConnectionStatus, EthernetStatus,
    CellularStatus, Co2, Pressure, Motion, DeskOccupancy,
]



from __future__ import annotations

from typing import Optional, Any

import disruptive
import disruptive.requests as dtrequests
from disruptive.resources.device import Device


class Emulator():
    """
    Contains staticmethods for the emulator resource.
    Used for namespacing only and thus does not have a constructor

    """

    @staticmethod
    def create_device(project_id: str,
                      device_type: str,
                      display_name: Optional[str] = None,
                      labels: dict[str, str] = {},
                      **kwargs: Any,
                      ) -> Device:
        """
        Create a new emulated device with specified type and project.

        Parameters
        ----------
        project_id : str
            Unique ID of the target project.
        device_type : str
            Specifies which device
            :ref:`type <device_type_constants>` to create.
        display_name : str, optional
            Provides a display name to the device.
        labels : dict[str, str], optional
            Set labels for the new device.
            Format: {'key': 'value'}.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        device : Device
            Object representing the created emulated device.

        Examples
        --------
        >>> # Create a new emulated humidity sensor.
        >>> device = dt.Emulator.create_device(
        ...     project_id='<PROJECT_ID>',
        ...     device_type=dt.Device.HUMIDITY,
        ...     display_name='new-humidity-sensor',
        ...     labels={
        ...         'room-number': '99',
        ...         'customer': 'x7',
        ...     },
        ... )

        """

        # Construct URL
        url = '/projects/{}/devices'.format(project_id)

        # Construct body dictionary.
        body: dict = dict()
        body['type'] = device_type
        body['labels'] = labels

        # Add display_name to labels dictionary in body.
        if display_name is not None:
            body['labels']['name'] = display_name

        # Return Device object of GET request response.
        return Device(dtrequests.DTRequest.post(
            url=url,
            base_url=disruptive.emulator_base_url,
            body=body,
            **kwargs,
        ))

    @staticmethod
    def delete_device(device_id: str,
                      project_id: str,
                      **kwargs: Any,
                      ) -> None:
        """
        Deletes the specified emulated device.

        Parameters
        ----------
        device_id : str
            Specifies which device type to delete.
        project_id : str
            Unique ID of the target project.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Examples
        --------
        >>> # Delete an emulated device.
        >>> dt.Emulator.delete_device('<DEVICE_ID>', '<PROJECT_ID>')

        """

        # Construct URL
        url = '/projects/{}/devices/{}'.format(project_id, device_id)

        # Send DELETE request, but return nothing.
        dtrequests.DTRequest.delete(
            url=url,
            base_url=disruptive.emulator_base_url,
            **kwargs,
        )

    @staticmethod
    def publish_event(device_id: str,
                      project_id: str,
                      data: disruptive.events.Touch |
                      disruptive.events.Temperature |
                      disruptive.events.ObjectPresent |
                      disruptive.events.Humidity |
                      disruptive.events.ObjectPresentCount |
                      disruptive.events.TouchCount |
                      disruptive.events.WaterPresent |
                      disruptive.events.NetworkStatus |
                      disruptive.events.BatteryStatus |
                      disruptive.events.ConnectionStatus |
                      disruptive.events.EthernetStatus |
                      disruptive.events.CellularStatus |
                      disruptive.events.Co2 |
                      disruptive.events.Pressure |
                      disruptive.events.Motion |
                      disruptive.events.DeskOccupancy |
                      disruptive.events.Contact |
                      disruptive.events.ProbeWireStatus,
                      **kwargs: Any,
                      ) -> None:
        """
        From the specified device, publish an event of the given type.

        Parameters
        ----------
        device_id : str
            Unique ID of the target device.
        project_id : str
            Unique ID of the target project.
        data : :ref:`Event Data <eventdata>`
            An object representing the event data to be published.
            Can be any of the listed Event Data classes.
            :code:`labelsChanged` is not supported
            when publishing emulated events.
            The chosen Event Data must be supported by the device.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Examples
        --------
        >>> # Publish a Humidity event from an emulated device.
        >>> dt.Emulator.publish_event(
        ...     device_id='<DEVICE_ID>',
        ...     project_id='<PROJECT_ID>',
        ...     data=dt.events.Humidity(
        ...         celsius=23,
        ...         humidity=76,
        ...     ),
        ... )

        """

        # Construct URL
        url = '/projects/{}/devices/{}:publish'.format(project_id, device_id)

        # Send POST request, but return nothing.
        dtrequests.DTRequest.post(
            url=url,
            base_url=disruptive.emulator_base_url,
            body={data.event_type: data._raw},
            **kwargs,
        )



# Metadata
__version__ = '1.6.4'

# If set, logs of chosen level and higher are printed to console.
# Default value None results in no logs at any level.
log_level = None

# REST API base URLs of which all endpoints are an expansion.
base_url = 'https://api.disruptive-technologies.com/v2'
emulator_base_url = 'https://emulator.disruptive-technologies.com/v2'

# If a request response contains an error for which a series of retries is
# worth considering, these variable determine how long to wait without an
# answer, and how many times the package should retry before raising an error.
request_timeout = 3  # seconds
request_attempts = 3  # attempts

# Authentication scheme.
from disruptive.authentication import Auth as Auth  # noqa

# Initialize package with environment variables authentication scheme.
default_auth = Auth.init()

# Resources.
from disruptive.resources.device import Device as Device  # noqa
from disruptive.resources.data_connector import DataConnector as DataConnector  # noqa
from disruptive.resources.organization import Organization as Organization  # noqa
from disruptive.resources.project import Project as Project  # noqa
from disruptive.resources.stream import Stream as Stream  # noqa
from disruptive.resources.eventhistory import EventHistory as EventHistory  # noqa
from disruptive.resources.service_account import ServiceAccount as ServiceAccount  # noqa
from disruptive.resources.service_account import Key as Key  # noqa
from disruptive.resources.role import Role as Role  # noqa
from disruptive.resources.emulator import Emulator as Emulator  # noqa
from disruptive.resources.claim import Claim as Claim  # noqa
from disruptive.outputs import Member as Member  # noqa

# Additional helper modules.
from disruptive import events as events  # noqa
from disruptive import outputs as outputs  # noqa
from disruptive import errors as errors  # noqa
from disruptive import logging as logging  # noqa



from __future__ import annotations

from typing import Any, Optional, List, Tuple

import disruptive.outputs as dtoutputs
import disruptive.requests as dtrequests
import disruptive.errors as dterrors


class Claim(dtoutputs.OutputBase):
    """
    Namespacing for claim methods.

    Attributes
    ----------
    type : str
        Whether or not the claim is for a KIT or DEVICE.
    claimed_item : Claim.ClaimKit | Claim.ClaimDevice
        An object representing the kit or device claim.

    """

    KIT = 'KIT'
    DEVICE = 'DEVICE'
    CLAIM_ITEMS = [KIT, DEVICE]

    def __init__(self, claim: dict) -> None:
        """
        Constructs the Claim object by unpacking raw claim response.

        Parameters
        ----------
        claim : dict
            Unmodified claim response dictionary.

        """

        # Inherit from OutputBase parent.
        dtoutputs.OutputBase.__init__(self, claim)

        # Unpack attributes from dictionary.
        self.type: str = claim['type']
        self.claimed_item: Claim.ClaimKit | Claim.ClaimDevice = \
            self._resolve_type(claim)

    @classmethod
    def claim_info(cls,
                   identifier: str,
                   organization_id: Optional[str] = None,
                   **kwargs: Any,
                   ) -> Claim:
        """
        Get claim information for either a device
        or a kit by looking up an identifier.

        For sensors, the identifier can be found in either
        the QR code or printed directly on the sensor.
        For kits, the identifier can be found both as text
        and a QR code printed on the box.

        For more information, see the
        `REST API Claim Info Documentation <https://developer.disruptive-
        technologies.com/api#/Claiming%20Devices%20%26%20Kits/
        get_claim_info>`_.

        Parameters
        ----------
        identifier : str
            Unique identifier of the target Kit or Device.
        organization_id : str, optional
            The identifier of the organization that will claim the
            devices or kits.

        Returns
        -------
        claim : Claim
            Claim information about the requested Kit or Device.

        Raises
        ------
        TypeError
            If provided identifier is not str type.

        Examples
        --------
        >>> # Get claim information about a kit.
        >>> claim = dt.Claim.claim_info('<KIT_ID>')

        >>> # Get claim information about a device.
        >>> claim = dt.Claim.claim_info('<DEVICE_ID>')

        """

        if not isinstance(identifier, str):
            raise TypeError(f'Identifier must be str, got {type(identifier)}.')

        url = f'/claimInfo?identifier={identifier}'

        # Add organization resource name to url if provided.
        if organization_id is not None:
            url += f'&organization=organizations/{organization_id}'

        return cls(dtrequests.DTRequest.get(url, **kwargs))

    @staticmethod
    def claim(target_project_id: str,
              kit_ids: Optional[List[str]] = None,
              device_ids: Optional[List[str]] = None,
              dry_run: bool = True,
              **kwargs: Any,
              ) -> Tuple[List[Claim.ClaimDevice], List[Exception]]:
        """
        Claim multiple kits and/or devices to your project.

        Claiming a kit/device does two things. It starts the device
        subscriptions, where any pre-paid period will be activated.
        It also adds the devices to your project.

        For more information, see the
        `REST API Claim Documentation <https://developer.disruptive-
        technologies.com/api#/Claiming%20Devices%20%26%20Kits/
        post_projects__project__devices_claim>`_.

        Parameters
        ----------
        target_project_id : str
            Unique identifier of project into which you wish to claim.
        kit_ids : list[str], optional
            List of unique kit IDs to claim.
        device_ids : list[str], optional
            List of unique device IDs to claim.
        dry_run : bool, optional
            Default True.
            Test your claim request during development.
            No kits or devices will be claimed while True.
            Set to False in production or no devices will be claimed.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        devices : list[Claim.ClaimDevice]
            List of successfully claimed devices.
        errors : list[ClaimError]
            List of errors that occured during claiming.
            If list is empty, all devices were claimed successfully.

        Examples
        --------
        >>> # Do a dry_run where you simulate claiming a kit.
        >>> devices, errors = dt.Claim.claim(
        ...     target_project_id='<TARGET_PROJECT_ID>',
        ...     kit_ids=['<KIT_ID>'],
        ... )

        >>> # Do a dry_run where you simulate claiming three devices.
        >>> devices, errors = dt.Claim.claim(
        ...     target_project_id='<TARGET_PROJECT_ID>',
        ...     device_ids=[
        ...         '<DEVICE_ID_1>',
        ...         '<DEVICE_ID_2>',
        ...         '<DEVICE_ID_3>',
        ...     ],
        ... )

        >>> # Turn off dry_run and claim 2 kits and 3 devices.
        >>> # This action is not reversible.
        >>> devices, errors = dt.Claim.claim(
        ...     target_project_id='<TARGET_PROJECT_ID>',
        ...     kit_ids=[
        ...         '<KIT_ID_1>',
        ...         '<KIT_ID_2>',
        ...     ],
        ...     device_ids=[
        ...         '<DEVICE_ID_1>',
        ...         '<DEVICE_ID_2>',
        ...         '<DEVICE_ID_3>',
        ...     ]
        ...     dry_run=False,
        ... )

        """

        url = f'/projects/{target_project_id}/devices:claim'
        url += f'?dryRun={str(dry_run).lower()}'

        body = {}
        if kit_ids is not None:
            body['kitIds'] = kit_ids
        if device_ids is not None:
            body['deviceIds'] = device_ids

        res = dtrequests.DTRequest.post(url, body=body, **kwargs)

        return (
            [Claim.ClaimDevice(d) for d in res['claimedDevices']],
            Claim._parse_claim_errors(res['claimErrors']),
        )

    @staticmethod
    def _parse_claim_errors(res_errors: dict) -> List[Exception]:
        errors: List[Exception] = []
        for error in res_errors['devices'] + res_errors['kits']:
            if error['code'] == 'ALREADY_CLAIMED':
                errors.append(dterrors.ClaimErrorDeviceAlreadyClaimed(error))
            elif error['code'] == 'NOT_FOUND' and 'deviceId' in error:
                errors.append(dterrors.ClaimErrorDeviceNotFound(error))
            elif error['code'] == 'NOT_FOUND' and 'kitId' in error:
                errors.append(dterrors.ClaimErrorKitNotFound(error))
            else:
                errors.append(dterrors.ClaimError(error))
        return errors

    def _resolve_type(self, claim: dict) -> Claim.ClaimKit | Claim.ClaimDevice:
        """
        Return either Kit or Device depending on type.

        Parameters
        ----------
        claim : dict
            Unmodified claim response dictionary.

        Returns
        -------
        claim_item : Claim.ClaimKit | Claim.ClaimDevice
            Either a Kit or Device object depending on type.

        """

        if claim['type'] == Claim.KIT:
            return Claim.ClaimKit(claim['kit'])
        elif claim['type'] == Claim.DEVICE:
            return Claim.ClaimDevice(claim['device'])
        else:
            raise KeyError(f'unknown claim type {claim["type"]}')

    class ClaimDevice(dtoutputs.OutputBase):
        """
        Namespacing type for a claimed device.

        Attributes
        ----------
        device_id : str
            Unique device identifier.
        device_type : str
            :ref:`Device type <device_type_constants>`.
        product_number : str
            The device product number.
        is_claimed : bool
            Whether or not the devices has been claimed.

        """

        def __init__(self, device: dict) -> None:
            """
            Constructs the claimed Device object from raw response.

            """

            # Inherit from OutputBase parent.
            dtoutputs.OutputBase.__init__(self, device)

            # Unpack attributes from raw response dictionary.
            self.device_id: str = device['deviceId']
            self.device_type: str = device['deviceType']
            self.product_number: str = device['productNumber']
            self.is_claimed: bool = device['isClaimed']

    class ClaimKit(dtoutputs.OutputBase):
        """
        Namespacing type for a claimed kit.

        Attributes
        ----------
        kit_id : str
            Unique kit identifier.
        display_name : str
            Human reabable kit name.
        devices : list[Claim.ClaimDevice]
            List of devices in the kit.

        """

        def __init__(self, kit: dict) -> None:
            """
            Constructs the claimed kit object from raw response.

            """

            # Inherit from OutputBase parent.
            dtoutputs.OutputBase.__init__(self, kit)

            # Unpack attributes from raw response dictionary.
            self.kit_id: str = kit['kitId']
            self.display_name: str = kit['displayName']
            self.devices: List[Claim.ClaimDevice] \
                = [Claim.ClaimDevice(d) for d in kit['devices']]



from __future__ import annotations

import os
import json
import time
import urllib.parse
from typing import Any

import jwt

from disruptive import requests as dtrequests, errors as dterrors


class _AuthRoutineBase(object):

    def __init__(self) -> None:
        # Set default attributes.
        self._expiration: int = 0
        self._token: str = ''

    def _has_expired(self) -> bool:
        """
        Evaluates whether the access token has expired.

        Returns
        -------
        has_expired : bool
            True if the access token has expired, otherwise False.

        """

        if time.time() > self._expiration:
            return True
        else:
            return False

    def get_token(self) -> str:
        """
        Returns the access token.
        If the token has expired, renew it.

        Returns
        -------
        token : str
            Access token added to the request header.

        """

        # Check expiration time.
        if self._has_expired():
            # Renew access token.
            self.refresh()

        return self._token

    def refresh(self) -> None:
        """
        This function does nothing and is overwritten in all
        child classes. It only exists for consistency purposes
        as it is called in get_token().

        """

        pass


class Unauthenticated(_AuthRoutineBase):

    def __init__(self) -> None:
        # Inherit parent class methods and attributes.
        super().__init__()

    def refresh(self) -> None:
        """
        If called, this function does nothing but raise an error as no
        authentication routine has been called to update the configuration
        variable, nor has an authentication object been provided.

        Raises
        ------
        Unauthorized
            If neither default_auth has been set nor the
            auth kwarg has been provided.

        """

        msg = 'Missing Service Account credentials.\n\n' \
            'Either set the following environment variables:\n\n' \
            '    DT_SERVICE_ACCOUNT_KEY_ID: Unique Service Account key ID.\n' \
            '    DT_SERVICE_ACCOUNT_SECRET: Unique Service Account secret.\n' \
            '    DT_SERVICE_ACCOUNT_EMAIL: Unique Service Account email.\n\n' \
            'or provide them programmatically:\n\n' \
            '    import disruptive as dt\n\n' \
            '    dt.default_auth = dt.Auth.service_account(\n' \
            '        key_id="<SERVICE_ACCOUNT_KEY_ID>",\n' \
            '        secret="<SERVICE_ACCOUNT_SECRET>",\n' \
            '        email="<SERVICE_ACCOUNT_EMAIL>",\n' \
            '    )\n\n' \
            'See https://developer.d21s.com/api/' \
            'libraries/python/client/authentication.html' \
            ' for more details.\n'

        raise dterrors.Unauthorized(msg)


class ServiceAccountAuth(_AuthRoutineBase):
    """
    Ensures that the access token is available and up-to-date.

    Attributes
    ----------
    token_endpoint : str
        URL to which the jwt is exchanged for an access token.

    """

    supported_algorithms = ['HS256']
    token_endpoint = 'https://identity.'\
        'disruptive-technologies.com/oauth2/token'

    def __init__(self, key_id: str, secret: str, email: str):
        # Inherit parent class methods and attributes.
        super().__init__()

        # Set parameter attributes.
        self._key_id: str = key_id
        self._secret: str = secret
        self._email: str = email

        # Default to HS256 algorithm.
        self._algorithm = self.supported_algorithms[0]

    @property
    def key_id(self) -> str:
        return self._key_id

    @property
    def secret(self) -> str:
        return self._secret

    @property
    def email(self) -> str:
        return self._email

    def __repr__(self) -> str:
        return '{}.{}({}, {}, {})'.format(
            self.__class__.__module__,
            self.__class__.__name__,
            repr(self.key_id),
            repr(self.secret),
            repr(self.email),
        )

    @property
    def algorithm(self) -> str:
        return self._algorithm

    @algorithm.setter
    def algorithm(self, algorithm: str) -> None:
        if algorithm in self.supported_algorithms:
            self._algorithm = algorithm
        else:
            raise dterrors.ConfigurationError(
                f'unsupported algorithm {algorithm}'
            )

    @classmethod
    def from_credentials_file(cls, credentials: dict) -> ServiceAccountAuth:
        for key in ['keyId', 'secret', 'email', 'algorithm', 'tokenEndpoint']:
            if key not in credentials['serviceAccount']:
                raise dterrors.ConfigurationError(
                    f'Invalid credentials file. Missing field "{key}".'
                )

        cfg = credentials['serviceAccount']
        auth_obj = cls(
            key_id=cfg['keyId'],
            secret=cfg['secret'],
            email=cfg['email'],
        )
        auth_obj.algorithm = cfg['algorithm']
        auth_obj.token_endpoint = cfg['tokenEndpoint']

        return auth_obj

    def refresh(self) -> None:
        """
        Refreshes the access token.

        This first exchanges the JWT for an access token, then updates
        the expiration and token attributes with the response.

        """

        response: dict = self._get_access_token()
        self._expiration = time.time() + response['expires_in']
        self._token = 'Bearer {}'.format(response['access_token'])

    def _get_access_token(self) -> dict:
        """
        Constructs and exchanges the JWT for an access token.

        Returns
        -------
        response : dict
            Dictionary containing expiration and the token itself.

        Raises
        ------
        BadRequest
            If the provided credentials could not be used for authentication.

        """

        # Construct the JWT header.
        jwt_headers: dict[str, str] = {
            'alg': self.algorithm,
            'kid': self.key_id,
        }

        # Construct the JWT payload.
        jwt_payload: dict[str, Any] = {
            'iat': int(time.time()),         # current unixtime
            'exp': int(time.time()) + 3600,  # expiration unixtime
            'aud': self.token_endpoint,
            'iss': self.email,
        }

        # Sign and encode JWT with the secret.
        encoded_jwt: str = jwt.encode(
            payload=jwt_payload,
            key=self.secret,
            algorithm=self.algorithm,
            headers=jwt_headers,
        )

        # Prepare HTTP POST request data.
        # Note: The requests package applies Form URL-Encoding by default.
        request_data: str = urllib.parse.urlencode({
            'assertion': encoded_jwt,
            'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer'
        })

        # Exchange the JWT for an access token.
        try:
            access_token_response: dict = dtrequests.DTRequest.post(
                url='',
                base_url=self.token_endpoint,
                data=request_data,
                headers={'Content-Type': 'application/x-www-form-urlencoded'},
                skip_auth=True,
            )
        except dterrors.BadRequest:
            # Re-raise exception with more specific information.
            raise dterrors.Unauthorized(
                'Could not authenticate with the provided credentials.\n\n'
                'Read more: https://developer.d21s.com/docs/authentication'
                '/oauth2#common-errors'
            )

        # Return the access token in the request.
        return access_token_response


def _service_account_env_vars() -> Unauthenticated | ServiceAccountAuth:
    key_id = os.getenv('DT_SERVICE_ACCOUNT_KEY_ID', '')
    secret = os.getenv('DT_SERVICE_ACCOUNT_SECRET', '')
    email = os.getenv('DT_SERVICE_ACCOUNT_EMAIL', '')

    if '' in [key_id, secret, email]:
        return Unauthenticated()
    else:
        return Auth.service_account(key_id, secret, email)


def _credentials_file() -> Unauthenticated | ServiceAccountAuth:
    file_path = os.getenv('DT_CREDENTIALS_FILE')
    if file_path is not None:
        if not os.path.exists(file_path):
            msg = 'Missing credentials file.\n\n' \
                'Environment variable DT_CREDENTIALS_FILE is set, but' \
                ' no file found at target path.\n' \
                f'{file_path}'
            raise FileNotFoundError(msg)
        with open(file_path, 'r') as f:
            credentials = json.load(f)

        if 'serviceAccount' in credentials:
            return ServiceAccountAuth.from_credentials_file(credentials)

    return Unauthenticated()


class Auth():
    """
    Authenticates the API using a factory design pattern.
    The Auth class itself is only for namespacing purposes.

    Depending on the classmethod called, an instance of some
    authentication routine, like ServiceAccountAuth, is returned.

    If no classmethod has been called, and the API is configured with
    an instance of the Auth class itself, exceptions will be raised
    with indicators on how to properly authenticate.

    """

    @staticmethod
    def init() -> Unauthenticated | ServiceAccountAuth:
        for method in [_credentials_file, _service_account_env_vars]:
            auth = method()
            if not isinstance(auth, Unauthenticated):
                return auth

        return Unauthenticated()

    @staticmethod
    def unauthenticated() -> Unauthenticated:
        return Unauthenticated()

    @classmethod
    def service_account(cls,
                        key_id: str,
                        secret: str,
                        email: str,
                        ) -> ServiceAccountAuth:
        """
        This method uses an OAuth2 authentication flow. With the provided
        credentials, a `JWT <https://jwt.io/>`_ is created and exchanged for
        an access token.

        Parameters
        ----------
        key_id : str
            Unique Service Account key ID.
        secret : str
            Service Account secret.
        email : str
            Unique Service Account email address.

        Returns
        -------
        auth : ServiceAccountAuth
            Object to initialize and maintain authentication to the REST API.

        Examples
        --------
        >>> # Authenticate using Service Account credentials.
        >>> dt.default_auth = dt.Auth.service_account(
        ...     key_id="<SERVICE_ACCOUNT_KEY_ID>",
        ...     secret="<SERVICE_ACCOUNT_KEY_ID>",
        ...     email="<SERVICE_ACCOUNT_KEY_ID>",
        ... )

        """

        # Check that credentials are populated strings.
        cls._verify_str_credentials({
            'key_id': key_id,
            'secret': secret,
            'email': email,
        })

        return ServiceAccountAuth(key_id, secret, email)

    @staticmethod
    def _verify_str_credentials(credentials: dict) -> None:
        """
        Verifies that the provided credentials are strings.

        This check is added as people use environment variables, but
        if for instance os.environ.get() does not find one, it silently
        returns None. It's better to just check for it early.

        Parameters
        ----------
        credentials : dict
            Credentials used to authenticate the REST API.

        """

        for key in credentials:
            # Verify credential is type string.
            if isinstance(credentials[key], str):
                # Raise ConfigurationError if string is empty.
                # This typically happens is credentials are fetched from
                # the environment with a fallback to an empty string.
                if len(credentials[key]) == 0:
                    raise dterrors.ConfigurationError(
                        'Authentication credential <{}> is'
                        ' empty string.'.format(key)
                    )

            # If not, raise TypeError.
            else:
                raise dterrors._raise_builtin(
                    TypeError,
                    'Authentication credential <{}> got type <{}>. '
                    'Expected <str>.'.format(
                        key, type(credentials[key]).__name__
                    )
                )



from disruptive.events.events import Event as Event  # noqa
from disruptive.events.events import Touch as Touch  # noqa
from disruptive.events.events import Temperature as Temperature  # noqa
from disruptive.events.events import TemperatureSample as TemperatureSample  # noqa
from disruptive.events.events import ObjectPresent as ObjectPresent  # noqa
from disruptive.events.events import Humidity as Humidity  # noqa
from disruptive.events.events import ObjectPresentCount as ObjectPresentCount  # noqa
from disruptive.events.events import TouchCount as TouchCount # noqa
from disruptive.events.events import WaterPresent as WaterPresent # noqa
from disruptive.events.events import NetworkStatus as NetworkStatus  # noqa
from disruptive.events.events import NetworkStatusCloudConnector as NetworkStatusCloudConnector  # noqa
from disruptive.events.events import BatteryStatus as BatteryStatus  # noqa
from disruptive.events.events import LabelsChanged as LabelsChanged # noqa
from disruptive.events.events import ConnectionStatus as ConnectionStatus # noqa
from disruptive.events.events import EthernetStatus as EthernetStatus  # noqa
from disruptive.events.events import CellularStatus as CellularStatus # noqa
from disruptive.events.events import Co2 as Co2  # noqa
from disruptive.events.events import Pressure as Pressure  # noqa
from disruptive.events.events import Motion as Motion  # noqa
from disruptive.events.events import DeskOccupancy as DeskOccupancy  # noqa
from disruptive.events.events import Contact as Contact  # noqa
from disruptive.events.events import ProbeWireStatus as ProbeWireStatus  # noqa

# Module constants.
from disruptive.events.events import TOUCH as TOUCH  # noqa
from disruptive.events.events import TEMPERATURE as TEMPERATURE  # noqa
from disruptive.events.events import OBJECT_PRESENT as OBJECT_PRESENT  # noqa
from disruptive.events.events import HUMIDITY as HUMIDITY  # noqa
from disruptive.events.events import OBJECT_PRESENT_COUNT as OBJECT_PRESENT_COUNT  # noqa
from disruptive.events.events import TOUCH_COUNT as TOUCH_COUNT  # noqa
from disruptive.events.events import WATER_PRESENT as WATER_PRESENT  # noqa
from disruptive.events.events import NETWORK_STATUS as NETWORK_STATUS  # noqa
from disruptive.events.events import BATTERY_STATUS as BATTERY_STATUS  # noqa
from disruptive.events.events import LABELS_CHANGED as LABELS_CHANGED  # noqa
from disruptive.events.events import CONNECTION_STATUS as CONNECTION_STATUS  # noqa
from disruptive.events.events import ETHERNET_STATUS as ETHERNET_STATUS  # noqa
from disruptive.events.events import CELLULAR_STATUS as CELLULAR_STATUS # noqa
from disruptive.events.events import CO2 as CO2 # noqa
from disruptive.events.events import PRESSURE as PRESSURE  # noqa
from disruptive.events.events import MOTION as MOTION  # noqa
from disruptive.events.events import DESK_OCCUPANCY as DESK_OCCUPANCY  # noqa
from disruptive.events.events import CONTACT as CONTACT  # noqa
from disruptive.events.events import PROBE_WIRE_STATUS as PROBE_WIRE_STATUS  # noqa
from disruptive.events.events import EVENT_TYPES as EVENT_TYPES  # noqa



from __future__ import annotations

from typing import Generator, Optional, Any

import disruptive.requests as dtrequests
from disruptive.events.events import Event


class Stream():
    """
    Contains staticmethods for streaming events.
    Used for namespacing only and thus does not have a constructor

    """

    @staticmethod
    def event_stream(project_id: str,
                     device_ids: Optional[list[str]] = None,
                     label_filters: Optional[dict] = None,
                     device_types: Optional[list[str]] = None,
                     event_types: Optional[list[str]] = None,
                     **kwargs: Any,
                     ) -> Generator:
        """
        Stream events for one, multiple, or all device(s) in a project.

        Implements a basic retry-routine. If connection is lost, the stream
        will attempt to reconnect with an exponential backoff. Events that
        are published during reconnection are not accounted for.

        If you want to forward your data in a server-to-server
        integration, consider using Data Connectors for a simpler
        and more reliable service with an added at-least-once guarantee.

        Parameters
        ----------
        project_id : str
            Unique ID of the target project.
        device_ids : list[str], optional
            Only includes events from the specified device(s).
        label_filters : dict[str, Optional[str]], optional
            Filter devices by their labels. Takes
            the form :code:`{"key": "value"}`, or :code:`{"key": None}` to
            allow any label value.
        device_types : list[str], optional
            Only includes events from devices with specified
            :ref:`type(s) <device_type_constants>`.
        event_types : list[str], optional
            Only includes events of the specified :ref:`type(s) <event_types>`.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        stream : Generator
            A python Generator type that yields each new event in the stream.

        Examples
        --------
        >>> # Sream real-time events from all devices in a project.
        >>> for event in dt.Stream.event_stream('<PROJECT_ID>'):
        ...     print(event)

        >>> # Sream real-time temperature- and touch events
        >>> # only from temperature devices that have set
        >>> # the labels `room-number=99` or `group=red`.
        >>> stream = dt.Stream.event_stream(
        ...     project_id='<PROJECT_ID>',
        ...     label_filters={
        ...         'room-number': '99',
        ...         'group': 'red',
        ...     },
        ...     device_types=[
        ...         dt.Device.TEMPERATURE,
        ...     ],
        ...     event_types=[
        ...         dt.events.TOUCH,
        ...         dt.events.TEMPERATURE,
        ...     ],
        ... )
        >>>
        >>> for event in stream:
        ...     print(event)

        """

        # Construct parameters dictionary.
        params: dict = dict()
        if device_ids is not None:
            params['device_ids'] = device_ids
        if device_types is not None:
            params['device_types'] = device_types
        if label_filters is not None:
            params['label_filters'] = []
            for key in label_filters:
                if isinstance(label_filters[key], str):
                    string = key + '=' + label_filters[key]
                    params['label_filters'].append(string)
                else:
                    params['label_filters'].append(key)
        if event_types is not None:
            params['event_types'] = event_types

        # Relay generator output.
        url = '/projects/{}/devices:stream'.format(project_id)
        for event in dtrequests.DTRequest.stream(url, params=params, **kwargs):
            yield Event(event)



from __future__ import annotations

from typing import Any

import disruptive.requests as dtrequests
from disruptive.outputs import OutputBase


class Role(OutputBase):
    """
    Represents a role.

    When a role response is received, the content is
    unpacked and the related attributes are updated.

    Attributes
    ----------
    role : str
        Name of the role.
    display_name : str
        Display name of the role.
    description : str
        Description of what the role entails.
    permissions : list[str]
        List of permissions provided with the role.
    raw : dict[str, str]
        Unmodified API response JSON.

    """

    # Constants for the available roles.
    PROJECT_USER = 'project.user'
    PROJECT_DEVELOPER = 'project.developer'
    PROJECT_ADMIN = 'project.admin'
    ORGANIZATION_ADMIN = 'organization.admin'
    ROLES = [PROJECT_USER, PROJECT_DEVELOPER, PROJECT_ADMIN,
             ORGANIZATION_ADMIN]

    def __init__(self, role: dict) -> None:
        """
        Constructs the Role object by unpacking the raw role response.

        Parameters
        ----------
        role : dict
            Unmodified role response dictionary.

        """

        # Inherit from Response parent.
        OutputBase.__init__(self, role)

        # Unpack attributes from dictionary.
        self.role: str = role['name'].split('/')[-1]
        self.display_name: str = role['displayName']
        self.description: str = role['description']
        self.permissions: list[str] = role['permissions']

    @classmethod
    def get_role(cls,
                 role: str,
                 **kwargs: Any,
                 ) -> Role:
        """
        Gets a role specified by its name.

        Parameters
        ----------
        role : str
            :ref:`Name <role_types>` of the role.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        role : Role
            Object representing the specified :ref:`role <role_types>`.

        Examples
        --------
        >>> # Fetch information about the project.developer role.
        >>> role = dt.Role.get_role(dt.Role.PROJECT_DEVELOPER)

        """

        # Return list of Role objects.
        return cls(dtrequests.DTRequest.get(
            url='/roles/' + role,
            **kwargs,
        ))

    @classmethod
    def list_roles(cls, **kwargs: Any) -> list[Role]:
        """
        List all available roles.

        Parameters
        ----------
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        roles : list[Role]
            List of objects each representing a :ref:`role <role_types>`.

        Examples
        --------
        >>> # Fetch information about all available roles.
        >>> roles = dt.Role.list_roles()

        """

        # Return list of Role objects.
        response = dtrequests.DTRequest.paginated_get(
            url='/roles',
            pagination_key='roles',
            **kwargs,
        )
        return [cls(r) for r in response]



from __future__ import annotations

import sys
import time
import json
from typing import Optional, Any, Generator

import requests

import disruptive as dt
import disruptive.logging as dtlog
import disruptive.errors as dterrors


USER_AGENT = 'DisruptivePythonAPI/{} Python/{}'.format(
    dt.__version__,
    f'{sys.version_info.major}.{sys.version_info.minor}',
)


class DTRequest():

    def __init__(self, method: str, url: str, **kwargs: Any):
        # Set attributes from parameters.
        self.method = method
        self.url = url

        # Set default attributes, though many are updated in _unpack_kwargs().
        self.base_url = dt.base_url
        self.params: dict = dict()
        self.headers: dict = dict()
        self.body = None
        self.data = None
        self.request_timeout = dt.request_timeout
        self.request_attempts = dt.request_attempts

        # Unpack kwargs and set attributes thereafter.
        self._unpack_kwargs(**kwargs)

        # Check that provided arguments are valid.
        self._sanitize_arguments()

        # Construct full url from base and target endpoint.
        self.full_url = self.base_url + self.url

    def _unpack_kwargs(self, **kwargs: Any) -> None:
        if 'params' in kwargs:
            self.params = kwargs['params']
        if 'headers' in kwargs:
            self.headers = kwargs['headers']
        if 'body' in kwargs:
            self.body = kwargs['body']
        if 'data' in kwargs:
            self.data = kwargs['data']

        # Check if request_timeout is overriden.
        if 'request_timeout' in kwargs:
            self.request_timeout = kwargs['request_timeout']

        # Check if request_attempts is overriden.
        if 'request_attempts' in kwargs:
            self.request_attempts = kwargs['request_attempts']

        # Check if base_url is overriden.
        if 'base_url' in kwargs:
            self.base_url = kwargs['base_url']

        # Add authorization header to request except when explicitly otherwise.
        if 'skip_auth' not in kwargs or kwargs['skip_auth'] is False:
            # If provided, override the package-wide auth with provided object.
            if 'auth' in kwargs:
                self.headers['Authorization'] = kwargs['auth'].get_token()
            # If not, use package-wide auth object.
            else:
                self.headers['Authorization'] = dt.default_auth.get_token()

    def _sanitize_arguments(self) -> None:
        # Check that request_timeout > 0.
        if self.request_timeout <= 0:
            raise dterrors.ConfigurationError(
                'Configuration parameter request_timeout has value {}, but '
                'must be float greater than 0.'.format(self.request_timeout)
            )

        # Check that request_attempts > 0.
        if self.request_attempts <= 0:
            raise dterrors.ConfigurationError(
                'Configuration parameter request_attempts has value {}, but '
                'must be integer greater than 0.'.format(self.request_attempts)
            )

    def _request_wrapper(self,
                         method: str,
                         url: str,
                         params: dict,
                         headers: dict,
                         body: Optional[dict],
                         data: Optional[str],
                         timeout: int,
                         ) -> tuple[DTResponse, Any]:

        # Add custom user agent.
        headers['User-Agent'] = USER_AGENT

        # Define default response values.
        res = None

        # Attempt to send the request.
        try:
            # Use the requests package to send the request.
            res = requests.request(
                method=method,
                url=url,
                params=params,
                headers=headers,
                json=body,
                data=data,
                timeout=timeout,
                stream=False,
            )

            # Isolate the data of interest in the response.
            return DTResponse(res.json(), res.status_code, res.headers), None

        except requests.exceptions.RequestException as e:
            return DTResponse({}, None, {}), e
        except ValueError as e:
            # Requests' .json() fails when no json is returned (code 405).
            if res is None:
                return DTResponse({}, 0, {}), e
            else:
                return DTResponse({}, res.status_code, res.headers), e

    def _send_request(self, nth_attempt: int = 0) -> dict:
        """
        Combines all the information and sends a request.

        Parameters
        ----------
        nth_attempt : int
            Request attempt count.

        Returns
        -------
        data : dict
            Data contained in the response.

        """

        # Log the request.
        dtlog.debug('Request [{}] to {}.'.format(
            self.method,
            self.base_url + self.url
        ))

        res, req_error = self._request_wrapper(
            method=self.method,
            url=self.full_url,
            params=self.params,
            headers=self.headers,
            body=self.body,
            data=self.data,
            timeout=self.request_timeout,
        )

        # Log the response.
        dtlog.debug('Response [{}].'.format(
            res.status_code
        ))

        # If _request_wrapper raised an exception, the request failed.
        if req_error is not None:
            error, should_retry, sleeptime = dterrors.parse_request_error(
                req_error, res.data, nth_attempt
            )
        else:
            # Parse the status_code and select an appropriate error.
            # If there is any hope at all that a retry might resolve the error,
            # should_retry will be True. (eg. a 401).
            error, should_retry, sleeptime = dterrors.parse_api_status_code(
                res.status_code, res.data, res.headers, nth_attempt
            )

        # Check if retry is required.
        if should_retry and nth_attempt < self.request_attempts:
            dtlog.warning('Reconnecting in {}s.'.format(sleeptime))

            # Sleep if necessary.
            if sleeptime is not None:
                time.sleep(sleeptime)

            dtlog.info('Connection attempt {} of {}.'.format(
                nth_attempt+1,
                self.request_attempts,
            ))

            # Attempt the request again recursively, iterating counter.
            res.data = self._send_request(nth_attempt+1)

        else:
            # If set, raise the error chosen by dterrors.parse_error().
            if error is not None:
                raise error

        data: dict = res.data
        return data

    @classmethod
    def get(cls, url: str, **kwargs: Any) -> dict:
        req = cls('GET', url, **kwargs)
        response: dict = req._send_request()
        return response

    @classmethod
    def post(cls, url: str, **kwargs: Any) -> dict:
        req = cls('POST', url, **kwargs)
        response: dict = req._send_request()
        return response

    @classmethod
    def patch(cls, url: str, **kwargs: Any) -> dict:
        req = cls('PATCH', url, **kwargs)
        response: dict = req._send_request()
        return response

    @classmethod
    def delete(cls, url: str, **kwargs: Any) -> dict:
        req = cls('DELETE', url, **kwargs)
        response: dict = req._send_request()
        return response

    @classmethod
    def paginated_get(cls,
                      url: str,
                      pagination_key: str,
                      params: dict[str, str] = {},
                      **kwargs: Any,
                      ) -> list:
        # Initialize output list.
        results = []

        # Loop until paging has finished.
        while True:
            response = cls.get(url, params=params, **kwargs)
            results += response[pagination_key]

            if len(response['nextPageToken']) > 0:
                params['pageToken'] = response['nextPageToken']
            else:
                break

        return results

    @staticmethod
    def stream(url: str, **kwargs: Any) -> Generator:
        """
        Initialzed and returns a stream generator.

        Parameters
        ----------
        url : str
            API endpoint URL.

        """

        # Set ping constants.
        PING_INTERVAL = 10
        PING_JITTER = 2

        # Expand url with base_url.
        url = dt.base_url + url

        # Set error variable that if not None, raise it.
        error = None

        # Unpack kwargs.
        params = kwargs['params'] if 'params' in kwargs else {}
        headers = kwargs['headers'] if 'headers' in kwargs else {}
        if 'request_attempts' in kwargs:
            request_attempts = kwargs['request_attempts']
        else:
            request_attempts = dt.request_attempts

        # Add ping parameter to dictionary.
        params['ping_interval'] = str(PING_INTERVAL) + 's'

        # Add custom user agent.
        headers['User-Agent'] = USER_AGENT

        # Set up a simple catch-all retry policy.
        nth_attempt = 0
        while True:
            try:
                # Set the authorization header each retry in case we expire.
                if 'auth' in kwargs:
                    headers['Authorization'] = kwargs['auth'].get_token()
                else:
                    headers['Authorization'] = dt.default_auth.get_token()

                # Set up a stream connection.
                # Connection will timeout and reconnect if no single event
                # is received in an interval of ping_interval + ping_jitter.
                dtlog.info('Starting stream...')
                stream = requests.request(
                    method='GET',
                    url=url,
                    stream=True,
                    timeout=PING_INTERVAL + PING_JITTER,
                    params=params,
                    headers=headers,
                    json=None,
                    data=None,
                )

                if stream.encoding is None:
                    stream.encoding = 'utf-8'

                # Iterate through the events as they come in (one per line).
                for line in stream.iter_lines(decode_unicode=True):
                    # Decode the response payload and break on error.
                    payload = json.loads(line)
                    if 'result' in payload:
                        # Reset retry counter.
                        nth_attempt = 0

                        # Check for ping event.
                        event = payload['result']['event']
                        if event['eventType'] == 'ping':
                            dtlog.debug('Ping received.')
                            continue

                        # Yield event to generator.
                        yield event

                    elif 'error' in payload:
                        error, _, _ = dterrors.parse_api_status_code(
                            payload['error']['code'],
                            payload, None, 0
                        )
                        raise error

                    else:
                        raise dterrors.UnknownError(payload)

                # If the stream finished, but without an error, break the loop.
                msg = 'Stream ended without an error.'
                raise dterrors.ConnectionError(msg)

            except KeyboardInterrupt:
                break

            except dterrors.DTApiError as e:
                # Except for Unauthorized, retry all DTApiErrors.
                if isinstance(e, dterrors.Unauthorized):
                    raise e
                elif nth_attempt < request_attempts:
                    sleeptime = nth_attempt**2

                    dtlog.warning('Reconnecting in {}s.'.format(sleeptime))

                    # Exponential backoff in sleep time.
                    time.sleep(sleeptime)

                    # Iterate attempt counter.
                    nth_attempt += 1
                    dtlog.info('Connection attempt {} of {}.'.format(
                        nth_attempt,
                        request_attempts,
                    ))
                else:
                    # To avoid printing the entire chain of re-raised
                    # exceptions, limit the traceback.
                    sys.tracebacklimit = 0
                    raise e

            except requests.exceptions.RequestException as e:
                # ConnectionErrors should always be retried.
                result = dterrors.parse_request_error(e, {}, nth_attempt)
                error, should_retry, sleeptime = result

                # Print the error and try again up to max_request_attempts.
                if nth_attempt < request_attempts and should_retry:
                    dtlog.warning('Reconnecting in {}s.'.format(sleeptime))

                    # Exponential backoff in sleep time.
                    time.sleep(sleeptime)

                    # Iterate attempt counter.
                    nth_attempt += 1
                    dtlog.info('Connection attempt {} of {}.'.format(
                        nth_attempt,
                        request_attempts,
                    ))

                else:
                    # To avoid printing the entire chain of re-raised
                    # exceptions, limit the traceback.
                    sys.tracebacklimit = 0
                    raise error from e


class DTResponse():

    def __init__(self,
                 data: dict,
                 status_code: Optional[int],
                 headers: Any,
                 ):

        self.data = data
        self.status_code = status_code
        self.headers = headers



from __future__ import annotations

from typing import Optional, Any
from datetime import datetime

import disruptive
import disruptive.requests as dtrequests
import disruptive.transforms as dttrans
from disruptive.events.events import Event


class EventHistory(list):
    """
    Namespacing type of event history methods.
    Inherits list, with some extra functionality.

    """

    @staticmethod
    def list_events(device_id: str,
                    project_id: str,
                    event_types: Optional[list[str]] = None,
                    start_time: Optional[str | datetime] = None,
                    end_time: Optional[str | datetime] = None,
                    **kwargs: Any,
                    ) -> EventHistory:
        """
        Get the event history for a single device.

        Parameters
        ----------
        device_id : str
            Unique ID of the target device.
        project_id : str
            Unique ID of the target project.
        event_types : list[str], optional
            If provided, only the specified
            :ref:`event types <event_types>` are fetched.
        start_time : str, datetime, optional
            Specifies from when event history is fetched.
            Defaults to 24 hours ago.
        end_time : str, datetime, optional
            Specified until when event history is fetched.
            Defaults to now.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        events : EventHistory[Event]
            A list of all events fetched by the call.

        Examples
        --------
        >>> # Fetch all events in the last 24h for a device.
        >>> events = dt.EventHistory.list_events(
        ...     device_id='<DEVICE_ID>',
        ...     project_id='<PROJECT_ID',
        ... )

        >>> # Fetch all touch- and objectPresent events
        >>> # for a device in the last 7 days.
        >>> events = dt.EventHistory.list_events(
        ...     device_id=DEVICE_1,
        ...     project_id=PROJECT_ID,
        ...     event_types=[
        ...         dt.events.TOUCH,
        ...         dt.events.OBJECT_PRESENT,
        ...     ],
        ...     start_time=datetime.utcnow() - timedelta(7),
        ... )

        """

        # Construct URL.
        url = '/projects/{}/devices/{}/events'.format(project_id, device_id)

        # Construct parameters dictionary.
        params: dict = dict()
        if event_types is not None:
            params['eventTypes'] = event_types

        # Sanitize timestamps as they must be iso8601 format.
        start_time_iso8601 = dttrans.to_iso8601(start_time)
        if start_time_iso8601 is not None:
            params['startTime'] = start_time_iso8601
        end_time_iso8601 = dttrans.to_iso8601(end_time)
        if end_time_iso8601 is not None:
            params['endTime'] = end_time_iso8601

        # Send paginated GET request.
        res = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='events',
            params=params,
            **kwargs,
        )

        # Return list of Event objects of paginated GET response.
        return EventHistory(Event.from_mixed_list(res))

    def to_dataframe(self) -> Any:
        """
        Experimental function to convert a list of events to DataFrame.
        Requires the installation of additional packages.
        >> pip install disruptive[extra]

        The `pandas` package is not, and will not, be a dependency of
        the core `disruptive` package. However, supporting DataFrames is a
        significant enough convenience that we're experimenting with it here.
        May be removed in the future if we decide that this is not a good idea.

        The columns `device_id`, `event_id`, and `event_type` are static, then
        one additional column is concatenated for every eventData field.

        Returns
        -------
        df : pandas.DataFrame
            DataFrame of all events in response.

        Raises
        ------
        ModueNotFoundError
            If we pandas package is not installed.

        """

        try:
            import pandas  # type: ignore
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                'Missing package `pandas`.\n\n'
                'to_dataframe() requires additional third-party packages.\n'
                '>> pip install disruptive[extra]'
            )

        rows = []
        for event in self:
            base = {
                'device_id': event.device_id,
                'event_id': event.event_id,
                'event_type': event.event_type,
            }

            if event.event_type == disruptive.events.TEMPERATURE:
                rows += [{
                        **base,
                        **sample.raw,
                    } for sample in event.data.samples
                ]
            else:
                rows.append({**base, **event.data.raw})

        df = pandas.json_normalize(
            rows, None, ['device_id', 'event_id', 'event_type'],
            errors='ignore',
        )

        # Convert columns headers from camelCase to snake_case for consistency.
        map = {name: dttrans.camel_to_snake_case(name) for name in df.columns}
        df = df.rename(columns=map)

        return df



from __future__ import annotations

import logging
from datetime import datetime

import disruptive
import disruptive.errors as dterrors

DEBUG = 'DEBUG'
INFO = 'INFO'
WARNING = 'WARNING'
ERROR = 'ERROR'
CRITICAL = 'CRITICAL'
LOG_LEVELS = [DEBUG, INFO, WARNING, ERROR, CRITICAL]

# Fetch the disruptive logger, but with disabled output.
logger = logging.getLogger('disruptive')
logger.setLevel(99)


def debug(msg: str | dict) -> None:
    if _log_flag_exceeds(DEBUG):
        _fmt_log(msg, DEBUG)
    logger.debug(msg)


def info(msg: str | dict) -> None:
    if _log_flag_exceeds(INFO):
        _fmt_log(msg, INFO)
    logger.info(msg)


def warning(msg: str | dict) -> None:
    if _log_flag_exceeds(WARNING):
        _fmt_log(msg, WARNING)
    logger.warning(msg)


def error(msg: str | dict) -> None:
    if _log_flag_exceeds(ERROR):
        _fmt_log(msg, ERROR)
    logger.error(msg)


def critical(msg: str | dict) -> None:
    if _log_flag_exceeds(CRITICAL):
        _fmt_log(msg, CRITICAL)
    logger.critical(msg)


def _log_flag_exceeds(level: str) -> bool:
    # If not string, never True.
    if not isinstance(disruptive.log_level, str):
        return False

    set_level = disruptive.log_level.upper()

    # Verify set value is valid.
    if set_level not in LOG_LEVELS:
        # As an invalid log_level has been provided, reset it
        # to default before raising the exception.
        disruptive.log_level = INFO

        msg = f'Invalid log_level {set_level}.\n' \
              f'Must be either of {LOG_LEVELS}.'
        raise dterrors.ConfigurationError(msg)

    # Check if level is exceeded.
    if LOG_LEVELS.index(level) >= LOG_LEVELS.index(set_level):
        return True
    else:
        return False


def _fmt_log(msg: str | dict, level: str) -> None:
    print(f'[{datetime.now().isoformat()}] {level:<8} - {msg}')
